"""
NRV-:class:`nrv_function` handling.
"""

from abc import abstractmethod

import numpy as np
from scipy.interpolate import CubicHermiteSpline, CubicSpline, Akima1DInterpolator, PchipInterpolator, make_interp_spline
from scipy.special import erf

from ..backend._file_handler import json_dump, json_load
from ..backend._log_interface import rise_warning, rise_error
from ..backend._NRV_Class import NRV_class, is_empty_iterable

#############################
## sigma functions classes ##
#############################
spy_interp1D_kind = {
    "linear":{"f":make_interp_spline, "kwgs":{"k":1}},
    "akima":{"f":Akima1DInterpolator, "kwgs":{"extrapolate":True}},
    "pchip":{"f":PchipInterpolator, "kwgs":{"extrapolate":True}},
    "slinear":{"f":make_interp_spline, "kwgs":{"k":1}},
    "quadratic":{"f":make_interp_spline, "kwgs":{"k":4}},
    "cubic":{"f":CubicSpline, "kwgs":{"extrapolate":True}},
    "previous":{"f":make_interp_spline, "kwgs":{"k":0}},
}


class nrv_function(NRV_class):
    """
    Class containg all comon method of fonction used in nrv
    """

    @abstractmethod
    def __init__(self):
        super().__init__()
        self.f_type = "nrv_function"
        self.dim = 0

    def __call__(self, *X):
        return self.call_method(*X)

    @staticmethod
    def call_method(self, X):
        return X

    def __compatible(self, b):
        if not callable(b):
            return True
        elif "dim" in self.__dir__() and "dim" in b.__dir__():
            if self.dim == b.dim:
                return True
        else:
            rise_warning("Type not compatible for operation", type(self), type(b))
            return False

    def __neg__(self):
        c = eval(self.f_type)()
        c.call_method = lambda *X: -self(*X)
        return c

    def __abs__(self):
        c = eval(self.f_type)()
        c.call_method = lambda *X: abs(self(*X))
        return c

    def __add__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: self(*X) + b(*X)
            else:
                c.call_method = lambda *X: self(*X) + b
            return c

    def __radd__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: b * (X) + self(*X)
            else:
                c.call_method = lambda *X: b + self(*X)
            return c

    def __sub__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: self(*X) - b(*X)
            else:
                c.call_method = lambda *X: self(*X) - b
            return c

    def __rsub__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: b(*X) - self(*X)
            else:
                c.call_method = lambda *X: b - self(*X)
            return c

    def __mul__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: self(*X) * b(*X)
            else:
                c.call_method = lambda *X: self(*X) * b
            return c

    def __rmul__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: b(*X) * self(*X)
            else:
                c.call_method = lambda *X: b * self(*X)
            return c


###############################################################
####################### 1D functions ##########################
###############################################################


class function_1D(nrv_function):
    """
    class containing function from IR to IR
    Such function can be call either on 1 value or on a ndarray (and applied on each value)
    """

    def __init__(self):
        super().__init__()
        self.f_type = "function_1D"
        self.dim = 1

    # @staticmethod
    def call_method(self, X):
        # Composition with N dim function
        if isinstance(X, nrv_function):
            c_type = "function_" + str(X.dim) + "D"
            c = eval(c_type)()
            c.call_method = lambda *x: self(X(*x))
            return c
        else:
            return None


class gaussian(function_1D):
    r"""
    gaussian function define as:

    .. math::

        f(x) = e^{-\frac{(x-\mu)^2}{2*\sigma^2}}
    """

    def __init__(self, mu=0, sigma=1):
        """ """
        super().__init__()
        self.mu = mu
        self.sigma = sigma

    def call_method(self, X):
        res = super().call_method(X)
        if res is None:
            res = np.exp(
                -(((X - self.mu) / self.sigma) ** 2) / 2
            )  # /(self.sigma*(2*np.pi)**0.5)
        return res


class gate(function_1D):
    def __init__(self, mu=0, sigma=1, kind="Rational", N=None):
        """ """
        super().__init__()
        # self.kind =
        self.mu = mu
        self.sigma = sigma
        self.N = N
        self.kind = kind

    def call_method(self, X):
        res = super().call_method(X)
        if res is None:
            X_eff = (X - self.mu) / self.sigma
            if self.N is None:
                res = (np.sign(X_eff + 0.5) - np.sign(X_eff - 0.5)) / 2
            elif self.kind.lower() == "rational":
                res = 1 / ((2 * X_eff) ** (2 * self.N) + 1)
            else:
                res = (
                    erf((X_eff + (0.5 * self.sigma) / self.mu) / self.N)
                    - erf((X_eff - (0.5 * self.sigma) / self.mu) / self.N)
                ) / 2
        return res


###############################################################
####################### 2D functions ##########################
###############################################################


class function_2D(nrv_function):
    """
    class containing function from IR^2 to IR
    Such function can be call either on 1 value or on a ndarray (and applied on each value)
    """

    def __init__(self):
        super().__init__()
        self.f_type = "function_2D"
        self.dim = 2

    @staticmethod
    def call_method(self, X):
        return 1


class ackley(function_2D):
    """
    Bi-dimensional Ackley function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array) -> np.array:
        return (
            -20 * np.exp(-0.2 * np.sqrt(0.5 * (x**2 + y**2)))
            - np.exp(0.5 * (np.cos(2 * np.pi * x) + np.cos(2 * np.pi * y)))
            + np.e
            + 20
        )


class beale(function_2D):
    """
    Bi-dimensional Beale function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array) -> np.array:
        return (
            (1.5 - x + x * y) ** 2
            + (2.25 - x + x * y**2) ** 2
            + (2.625 - x + x * y**3) ** 2
        )


class goldstein_price(function_2D):
    """
    Bi-dimensional Goldstein function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array) -> np.array:
        return (
            1
            + (x + y + 1) ** 2
            * (19 - 14 * x + 3 * x**2 - 14 * y + 6 * x * y + 3 * y**2)
        ) * (
            30
            + (2 * x - 3 * y) ** 2
            * (18 - 32 * x + 12 * x**2 + 48 * y - 36 * x * y + 27 * y**2)
        )


class booth(function_2D):
    """
    Bi-dimensional Booth function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array) -> np.array:
        return (x + 2 * y - 7) ** 2 + (2 * x + y - 5) ** 2


class bukin6(function_2D):
    """
    Bi-dimensional Booth function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array) -> np.array:
        return 100 * np.sqrt(np.abs(y - 0.01 * x**2)) + 0.01 * np.abs(x + 10)


###############################################################
####################### ND functions ##########################
###############################################################
class function_ND(nrv_function):
    """
    class containing function from IR^n to IR
    Such function can be call either on 1 value or on a ndarray (and applied on each value)
    """

    def __init__(self):
        super().__init__()
        self.f_type = "function_ND"
        self.dim = "N"

    @staticmethod
    def call_method(self, *X):
        return 1


class Id(function_ND):
    """ """

    def __init__(self):
        super().__init__()

    def call_method(self, *x: np.array) -> np.array:
        if len(x) == 1 and np.iterable(x[0]):
            x = x[0]
        return x


class rosenbock(function_ND):
    """
    Multi-dimensional Rosenbock function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, *x: np.array) -> np.array:
        if len(x) == 1 and np.iterable(x[0]):
            x = x[0]
        result = 0
        for i in range(1, len(x)):
            result += 100 * (x[i] - x[i - 1] ** 2) ** 2 + (1 - x[i - 1]) ** 2
        return result


class rastrigin(function_ND):
    """
    Multi-dimensional Rastrigin function
    """

    def __init__(self):
        super().__init__()

    def call_method(self, *x: np.array) -> np.array:
        if len(x) == 1 and np.iterable(x[0]):
            x = x[0]
        A = 10.0
        n = len(x)
        result = A * n
        for xi in x:
            result += xi**2 - A * np.cos(2 * np.pi * xi)
        return result


class sphere(function_ND):
    """
    Multi-dimensional Sphere function
    """

    def __init__(self, Xc=None):
        super().__init__()
        self.Xc = Xc
        if self.Xc is None:
            self.Xc = []

    def call_method(self, *X):
        if len(X) == 1 and np.iterable(X[0]):
            X = X[0]
        Nc = len(self.Xc)
        res = 0
        for i, xi in enumerate(X):
            if i < Nc:
                xc = self.Xc[i]
            else:
                xc = 0
            res += (xi - xc) ** 2
        return res**0.5


###########################################################
####################  cost_evaluation  #####################
###########################################################
class cost_evaluation(nrv_function):
    def __init__(self):
        super().__init__()
        self.f_type = "cost_evaluation"

    @staticmethod
    def call_method(self, static_sim) -> float:
        return 1

    def __call__(self, results, **kwargs) -> float:
        return self.call_method(results)


###############################################################
####################### interpolation ######################
###############################################################
class nrv_interp(nrv_function):
    """
    Interpolator based on scipy.interpolate

    Parameters
    ----------
    X_values    : (N,) array_like
        A 1-D array of real values
    Y_values    :(…,N,…) array_like
        A N-D array of real values. The length of y along the interpolation
        axis must be equal to the length of x.
    kind        : str
        kind of interpolation which should be used, by default "linear" possible values:
            - scipy interp1d kinds (see nrv.spy_interp1D_kind)
            - {'hermite', 'cardinal', 'catmull-rom'} for respective
              cubic spline interpollation
    dx
        if minimal distance between 2 consecutive points, by default 0.01 interpolator
        custumize intepolation function if None kind is use to set the interpolator,
        by default None
    dxdy
        if kind is 'hermite', derivative vector (see scipy.interpolate.CubicHermiteSpline),
        by default None
    scale
        if kind is 'cardinal', scale to use for derivative vector, by default None
    columns       : int or array_like
        columns on which apply the interpolation when class called
        if [] iterpolation done on the whole vector,by default []

    Returns
    -------
    output  :  tuple(3)
        (domain, cell_tag, facet_tag)
    """

    def __init__(
        self,
        X_values,
        Y_values,
        kind="linear",
        dx=0.01,
        interpolator=None,
        dxdy=None,
        scale=None,
        columns=[],
    ):
        super().__init__()
        self.f_type = "nrv_interp"
        # General parameters
        self.X_values = X_values
        self.Y_values = Y_values
        self.N_pts = len(Y_values)
        self.interpolator = interpolator
        self.kind = kind
        self.dx = dx

        # CubicHermiteSpline parameters
        self.dxdy = dxdy

        # Cardinal Spline parameters
        self.scale = scale

        self.columns = columns
        self.__set_interp()

    def __set_interp(self):
        """
        Internal use only, set interpolator
        """
        self.N_pts = len(self.X_values)
        self.__check_X_values()
        self.update_interpolator()

    def __check_X_values(self, dx=None):
        """ """
        S = self.X_values.argsort()
        self.X_values = self.X_values[S]
        self.Y_values = self.Y_values[S]
        if dx is not None:
            self.dx = dx
        for i, x in enumerate(self.X_values):
            if i > 0:
                while x < self.X_values[i - 1]:
                    x += dx
                self.X_values[i] = x

    def update_interpolator(self, kind=None, interpolator=None, dxdy=None, scale=None):
        """ """
        if kind is not None:
            self.kind = kind
        if interpolator is not None:
            self.interpolator = interpolator
        elif self.kind.lower() in spy_interp1D_kind:
            __interp_type = spy_interp1D_kind[self.kind.lower()]["f"]
            __kwgs = spy_interp1D_kind[self.kind.lower()]["kwgs"]
            self.interpolator = __interp_type(
                self.X_values, self.Y_values, **__kwgs
            )
        elif self.kind.lower() in ["hermite", "cardinal", "catmull-rom"]:
            if self.scale is None or self.kind == "catmull-rom":
                self.scale = 0.5
            if self.dxdy is None:
                self.dxdy = np.array(
                    [
                        (self.Y_values[1] - self.Y_values[0]) / self.X_values[1]
                        - self.X_values[0]
                    ]
                    + [
                        (self.Y_values[k + 2] - self.Y_values[k])
                        / (self.X_values[k + 2] - self.X_values[k])
                        for k in range(self.N_pts - 2)
                    ]
                    + [
                        (self.Y_values[-1] - self.Y_values[-2])
                        / (self.X_values[-1] - self.X_values[-2])
                    ]
                )
            if self.kind.lower() == "hermite" and dxdy is not None:
                self.dxdy = dxdy
                self.scale = 1
            if self.kind.lower() == "cardinal" and scale is not None:
                self.scale = scale

            self.interpolator = CubicHermiteSpline(
                self.X_values, self.Y_values, self.scale * self.dxdy
            )

    # Mathematical operations
    def __add__(self, b):
        if isinstance(b, nrv_interp):
            if np.allclose(b.X_values, self.X_values):
                Y = b.Y_values + self.Y_values
                # TO ADD: computation of dxdy when both are set
            else:
                rise_error(
                    "Not implemented: operations of nrv_iterp with different X_scale"
                )
        else:
            Y = self.Y_values + b
        return nrv_interp(
            X_values=self.X_values,
            Y_values=Y,
            interpolator=self.interpolator,
            kind=self.kind,
            dx=self.dx,
            dxdy=None,
            scale=self.scale,
            columns=self.columns,
        )

    def __mul__(self, b):
        if isinstance(b, nrv_interp):
            if np.allclose(b.X_values, self.X_values):
                Y = self.Y_values * b.Y_values
                # TO ADD: computation of dxdy when both are set
            else:
                rise_error(
                    "Not implemented: operations of nrv_iterp with different X_scale"
                )
        else:
            Y = self.Y_values * b
        return nrv_interp(
            X_values=self.X_values,
            Y_values=Y,
            interpolator=self.interpolator,
            kind=self.kind,
            dx=self.dx,
            dxdy=None,
            scale=self.scale,
            columns=self.columns,
        )

    def __truediv__(self, b):
        if isinstance(b, nrv_interp):
            if np.allclose(b.X_values, self.X_values):
                if np.isclose(b.Y_values, 0).any():
                    rise_error(ZeroDivisionError, "float division by zero in Python")
                Y = self.Y_values / b.Y_values
                # TO ADD: computation of dxdy when both are set
            else:
                rise_error(
                    NotImplementedError,
                    "operations of nrv_iterp with different X_scale",
                )
        else:
            Y = self.Y_values / b
        return nrv_interp(
            X_values=self.X_values,
            Y_values=Y,
            interpolator=self.interpolator,
            kind=self.kind,
            dx=self.dx,
            dxdy=None,
            scale=self.scale,
            columns=self.columns,
        )

    def __rtruediv__(self, b):
        if np.isclose(self.Y_values, 0).any():
            rise_error(ZeroDivisionError, " float division by zero in Python")
        if isinstance(b, nrv_interp):
            if np.allclose(b.X_values, self.X_values):
                Y = b.Y_values / self.Y_values
                # TO ADD: computation of dxdy when both are set
            else:
                rise_error(
                    NotImplementedError,
                    "operations of nrv_iterp with different X_scale",
                )
        else:
            Y = b / self.Y_values
        return nrv_interp(
            X_values=self.X_values,
            Y_values=Y,
            interpolator=self.interpolator,
            kind=self.kind,
            dx=self.dx,
            dxdy=self.dxdy,
            scale=self.scale,
            columns=self.columns,
        )

    def __radd__(self, b):
        return self.__add__(b)

    def __rmul__(self, b):
        return self.__mul__(b)

    def __sub__(self, b):
        return self.__add__(-b)

    def __rsub__(self, b):
        return self.__sub__(b)

    def __call__(self, X):
        if is_empty_iterable(self.columns):
            return self.interpolator(X)
        try:
            return self.interpolator(X[self.columns])
        except:
            rise_warning(
                "nrv_interpol: columns out of bound, intepolation done on the whole vector"
            )
            return self.interpolator(X)


class MeshCallBack(nrv_function):
    """ """

    def __init__(self, f=None, axis="x"):
        super().__init__()
        self.f_type = "MeshCallBack"
        self.f = None
        self.axis = axis

        self.set_function(f)

    def set_function(self, f=None):
        if f is None:
            self.f = lambda *x: 1
        elif callable(f):
            self.f = f
        elif isinstance(f, str):
            self.f = lambda x: eval(f)
        else:
            rise_warning(type(f), "Not recognized for MeshCallBack function")

    def __call__(self, dim, tag, x, y, z, lc):
        X = []
        if "x" in self.axis:
            X += [x]
        if "y" in self.axis:
            X += [y]
        if "z" in self.axis:
            X += [z]
        return lc * self.f(*X)



import numpy as np
import matplotlib.pyplot as plt
import csv

from ...backend._file_handler import rmv_ext
from ...backend._log_interface import rise_warning
from ...utils._nrv_function import nrv_interp


####################################################################
################# generate waveform functions ######################
####################################################################


## interpolation
def interpolate(
    y: np.ndarray,
    x: np.ndarray = [],
    scale=4,
    intertype="Spline",
    bounds=(0, 0),
    save=False,
    filename="interpolate.dat",
    save_scale=False,
    kwargs_interp={},
):
    """
    :meta private:

    Parameters
    ----------
    y : np.ndarray
        _description_
    x : np.ndarray, optional
        _description_, by default []
    scale : int, optional
        _description_, by default 4
    intertype : str, optional
        _description_, by default "Spline"
    bounds : tuple, optional
        _description_, by default (0, 0)
    save : bool, optional
        _description_, by default False
    filename : str, optional
        _description_, by default "interpolate.dat"
    save_scale : bool, optional
        _description_, by default False
    kwargs_interp : dict, optional
        _description_, by default {}

    Returns
    -------
    _type_
        _description_
    """

    y_scale = []

    if len(x) == 0:
        x = [i + 1 for i in range(len(y))]

    if type(scale) == float or type(scale) == int:
        x_scale = np.linspace(x[0], x[-1], int(scale))
    elif len(scale) >= 1:
        x_scale = scale
    else:
        x_scale = np.linspace(x[0], x[-1], (1 / 100) * (len(x) - 1) + 1)

    if intertype.lower() == "spline":
        intertype = "catmull-rom"
    y_interpol = nrv_interp(x, y, kind=intertype.lower(), **kwargs_interp)
    y_scale = y_interpol(x_scale)

    if bounds[0] != bounds[1]:
        for i in range(len(y_scale)):
            if y_scale[i] < min(bounds):
                y_scale[i] = min(bounds)

            if y_scale[i] > max(bounds):
                y_scale[i] = max(bounds)

    if save:
        np.set_printoptions(threshold=40000)
        if save_scale and filename[-3:] == "csv":
            np.savetxt(
                filename, np.transpose(np.array([x_scale, y_scale])), delimiter=","
            )
        else:
            y_str = np.array2string(y_scale, separator="\n")
            file = open(filename, "w")
            file.write(y_str[1:-2])
            file.close()
        np.set_printoptions(threshold=10)
    return y_scale


def interpolate_amp(
    position: np.ndarray,
    t_sim: float = 100,
    t_end: float = None,
    dt: float = 0.005,
    intertype: str = "Spline",
    bounds: tuple[float] = (0, 0),
    save: bool = False,
    filename: str = "interpolate_part.dat",
    save_scale: bool = False,
) -> np.ndarray:
    """
    genarte a waveform from a particle position using interpolate where the position values are the output waveform amplitudes at constant sample rate

    Parameters
    ----------
    position    : array
        particle position in n dimension output waveform amplitudes at regular times
    t_sim       : float
        simulation time (ms), by default 100
    dt       	: float
        time step of the simulation (ms), by default 0.005
    intertype   : str
        type of interpolation perform, by default 'Spline'
        type possibly:

            - 'Spline'                : Cubic spline interpolation
    bounds      : tupple
        limit range of the interpolation, if both equal no limit,by default (0,0)
    save        : bool
        save or not the output in a .dat file, by default False
    filename    : str
        name of the file on wich the output should be saved, by default 'interpolate_part.dat'

    Returns
    -------
    waveform     : np.ndarray
        result of the interpolation
    """
    if t_end is None:
        t_end = t_sim

    dim = len(position)
    time_particle = np.linspace(0, t_end, dim)
    scale = int(t_end / dt)

    waveform = interpolate(
        position,
        x=time_particle,
        scale=scale,
        intertype=intertype,
        bounds=bounds,
        save=save,
        filename=filename,
        save_scale=save_scale,
    )

    if t_end < t_sim:
        dif = int((t_sim - t_end) / dt)
        waveform = np.concatenate((waveform, np.ones(dif)))
    elif t_end > t_sim:
        waveform = waveform[: int(t_sim / dt) + 1]

    return waveform


def interpolate_Npts(
    position,
    t_sim: float = 100,
    dt: float = 0.005,
    amp_start: float = 0,
    amp_stop: float = 1,
    intertype: str = "Spline",
    bounds: tuple[float] = (0, 0),
    fixed_order: bool = False,
    t_end: float = None,
    t_shift: float = None,
    save: bool = False,
    fname: str = "interpolate_2pts.dat",
    plot: bool = False,
    save_scale: bool = False,
    generatefigure: bool = True,
    strict_bounds: bool = True,
    kwargs_interp: dict = {},
    **kwargs
):
    r"""
    genarte a waveform from a particle position using interpolate where the position
    values are the coordonnate of N points which should be reached by the output waveform

    Note
    ----
    If :math:`t_{i}` and :math:`I_{i}` are the time and amplitude of the :math:`ith` point the position vector :math:`\mathcal{X}` should be:

    .. math::

        \mathcal{X} = \begin{pmatrix} t_{1} & I_{1} & t_{2} & I_{2} & ... & t_{N} & I_{N}   \end{pmatrix}

    Parameters
    ----------
    position    : np.ndarray
        particle position in 2N dimensions with the coordonnate of the N points to interpolate.
    t_sim       : float
        simulation time (ms), by default 100
    dt          : float
        time step of the simulation (ms), by default 0.005
    amp_start   : float
        amplitude at the beginning of the interpolation
    amp_stop    : float
        amplitude at the end of the interpolation
    intertype   : str
        type of interpolation perform, by default 'Spline'
        type possibly:

            - 'Spline'                : Cubic spline interpolation
            - 'linear'
    bounds      : tupple
        limit range of the interpolation, if both equal no limit, by default (0,0)
    fixed_order         :bool
        fix the order of the points to interpolate
    t_end           :float
        optionnal, if not None, time of the stimulation at which the interpollation should reach amp_stop
    t_shift         :float
        optionnal, if not None, interpolation will be shifted of this time
    save        : bool
        save or not the output in a .dat file, by default False
    fname    : str
        name of the file on wich the output should be saved, by default 'interpolate_2pts.dat'
    strict_bounds    :bool
        if True values out of bound will be set to closer bound
    kwargs_interp   : dict
        kwargs to add to the interpollation

    Returns
    -------
    waveform     : np.ndarray
        waveform generated from position
    """

    n_dim = len(position)
    n_pts = n_dim // 2
    X = np.array(
        [[position[2 * k], position[2 * k + 1]] for k in range(n_pts)], dtype=float
    )

    # if odd number of dimention the last scalar is use to set t_end
    if n_dim % 2 == 1:
        t_end = max(position[-1], 2 * n_pts * dt)
        X[:, 0] *= t_end
    elif t_end is None:
        t_end = t_sim

    for i in range(n_pts):
        if X[i, 0] < dt:
            X[i, 0] += dt
        elif X[i, 0] > t_end - (n_pts * dt):
            X[i, 0] -= n_pts * dt
    if fixed_order:
        rise_warning(NotImplemented, "fixed_order not NotImplemented")
        fixed_order = False
    if fixed_order:
        pass
    else:
        I = np.argsort(X[:, 0])
        X = X[I]
        for i in range(n_pts):
            for j in range(i + 1, n_pts):
                if X[i, 0] + dt > X[j, 0]:
                    X[j, 0] = X[j, 0] + dt

    t = np.concatenate([[0], X[:, 0], [t_end]])
    x = np.concatenate([[amp_start], X[:, 1], [amp_stop]])

    I = np.argsort(t)
    t = t[I]
    x = x[I]

    if strict_bounds:
        bds = bounds
    else:
        bds = (0, 0)

    waveform = interpolate(
        y=x,
        x=t,
        scale=int(t_end / dt) + 1,
        intertype=intertype,
        bounds=bds,
        save=False,
        save_scale=save_scale,
        **kwargs_interp
    )

    if t_shift is not None:
        dif = int(t_shift / dt)
        waveform = np.concatenate((amp_start * np.ones(dif), waveform))
        t_end += t_shift
    if t_end < t_sim:
        dif = int((t_sim - t_end) / dt)
        waveform = np.concatenate((waveform, amp_stop * np.ones(dif)))
    elif t_end > t_sim:
        waveform = waveform[: int(t_sim / dt) + 1]
    if save or plot:
        T = np.linspace(0, t_sim, len(waveform))
        if generatefigure:
            plt.figure()
        plt.plot(T, waveform)
        plt.scatter(t, x)
        if save:
            plt.savefig(fname)
    return waveform


####################################################################
########################### savers #################################
####################################################################


def cost_position_saver(data, file_name="document.csv"):
    """
    Simple saver which can be used in a cost_function to save the cost
    and position in a .csv file (see .Optim.cost_function)

    Parameters
    ----------
    data        : dict
        dict containing the keys 'cost' and 'position'
    file_name:
        name of the saving file.
        NB: if missing, extension ".csv" will be add at the end of the file
    """
    save = [str(data["cost"])]
    position = data["position"]
    fname = rmv_ext(file_name) + ".csv"
    dim = len(position)
    for i in range(dim):
        save += [position[i]]
    with open(fname, "a", newline="") as fd:
        writer = csv.writer(fd)
        writer.writerow(save)



import os
import sys
import psutil

from ._GPU import getGPUs
from ._NRV_Singleton import NRV_singleton


# miscalleneous constants
Mem_KBytes = 1024
Mem_MBytes = Mem_KBytes**2
Mem_GBytes = Mem_KBytes**3
# Constants for sanity check
suported_platforms = [
    "linux",
    "darwin",
    "win32",
    "cygwin",
]
# python min vesion
NRV_PYTHON_VERSION = {
    "major": 3,
    "minor": 12,
    "patch": 0,
}

MIN_REQUIREMENTS = {
    "CPU_ncores": 1,
    "Memory_size": 8000
}


class MachineConfig(metaclass=NRV_singleton):
    """Singleton to get all machine properties related to computational performances.

    Parameters
    ----------
    memory_unit: int
        Dividor to express all memomry sizes in a given unit, 
        Constant already defined, can be:
        * Mem_KBytes (= 1024)
        * Mem_MBytes (= Mem_KBytes**2)
        * Mem_GBytes (= Mem_KBytes**3
    """
    def __init__(self, memory_unit=Mem_MBytes):
        self._explore_OS()
        self._explore_Python()
        self._explore_Float()
        self._explore_CPU()
        self._explore_GPU()
        self._explore_memory(unit=memory_unit)

    def __str__(self):
        current_str = f"""%%%%%%%%%%%%%%%%%%%%%%%%
%% INFORMATIONS ON OS %%
%%%%%%%%%%%%%%%%%%%%%%%%
OS platform: {self.OS_name}
OS version: {self.OS_version}

%%%%%%%%%%%%%%%%%%%%%%%%%
%% INFORMATIONS ON CPU %%
%%%%%%%%%%%%%%%%%%%%%%%%%
Number of available Cores: {self.CPU_ncores}
CPU number of bits: {self.CPU_nbits}
CPU architecture: {self.CPU_arch}

%%%%%%%%%%%%%%%%%%%%%%%%%
%% INFORMATIONS ON GPU %%
%%%%%%%%%%%%%%%%%%%%%%%%%
Number of available Cores: {self.GPU_ncores}
GPU total memory: {self.GPU_memory_total}
GPU used memory: {self.GPU_memory_used}
GPU free memory: {self.GPU_memory_free}
GPU memory load: {self.GPU_memory_load}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INFORMATIONS ON MEMORY %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
MEMORY total size: {self.MEMORY_size} {self.MEMORY_unit}
MEMORY available: {self.MEMORY_available} {self.MEMORY_unit}
MEMORY available in percent: {self.MEMORY_percent}
MEMORY used: {self.MEMORY_used} {self.MEMORY_unit}
MEMORY free: {self.MEMORY_free} {self.MEMORY_unit}
MEMORY active: {self.MEMORY_active} {self.MEMORY_unit}
MEMORY inactive: {self.MEMORY_inactive} {self.MEMORY_unit}
MEMORY wired: {self.MEMORY_wired} {self.MEMORY_unit}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INFORMATIONS ON PYTHON %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Python version: {self.Python_version_major}.{self.Python_version_minor}.{self.Python_version_micro}
Release level: {self.Python_version_releaselevel}
serial: {self.Python_version_serial}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% INFORMATION ON FLOAT REPRESENTATION %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Maximum number of decimal digits that can be faithfully represented in a float: {self.dig}
Difference between 1 and the next representable float: {self.epsilon}
Mantissa digits: {self.mant_dig}
Maximum representable finite floats: {self.max}
Maximum int e such that 10**e, that radix**(e-1), are representable: {self.max_10_exp}, {self.max_exp}
Minimum positive normalized float: {self.min}
Minimum int e such that 10**e, that radix**(e-1), is a normalized float{self.min_10_exp}, {self.min_exp}
Radix of exponent: {self.radix}
Rounding mode used for arithmetic operations: {self.rounds}
Float representation style: {self.float_repr_style}"""
        return current_str

    def _explore_OS(self):
        self.OS_name = sys.platform
        self.OS_version = os.uname().release

    def _explore_Python(self):
        # Python version
        self.Python_version_major = sys.version_info.major
        self.Python_version_minor = sys.version_info.minor
        self.Python_version_micro = sys.version_info.micro
        self.Python_version_releaselevel = sys.version_info.releaselevel
        self.Python_version_serial = sys.version_info.serial
        # imported modules from sys.modules, may be track versions

    def _explore_Float(self):
        self.dig = sys.float_info.dig
        self.epsilon = sys.float_info.epsilon
        self.mant_dig = sys.float_info.mant_dig
        self.max = sys.float_info.max
        self.max_10_exp = sys.float_info.max_10_exp
        self.max_exp = sys.float_info.max_exp
        self.min = sys.float_info.min
        self.min_10_exp = sys.float_info.min_10_exp
        self.min_exp = sys.float_info.min_exp
        self.radix = sys.float_info.radix
        self.rounds = sys.float_info.rounds
        self.float_repr_style = sys.float_repr_style

    def _explore_CPU(self):
        # get number of cores
        self.CPU_ncores = os.cpu_count()
        # get number of bits
        if sys.maxsize == 2**63 - 1:
            self.CPU_nbits = 64
        elif sys.maxsize == 2**31 - 1:
            self.CPU_nbits = 32
        else:
            print("this is really wierd...")
            self.CPU_nbits = 0.5
        self.CPU_arch = os.uname().machine

    def _explore_GPU(self):
        GPUs = getGPUs()
        self.GPU_ncores = len(GPUs)
        self.GPU_total_memory = 0
        self.GPU_memory_total = 0
        self.GPU_memory_used = 0
        self.GPU_memory_free = 0
        self.GPU_memory_load = 0
        if len(GPUs) > 0:
            for g in GPUs:
                self.GPU_memory_total += g.memoryTotal
                self.GPU_memory_used += g.memoryUsed
                self.GPU_memory_free += g.memoryFree
            self.GPU_memory_load = self.GPU_memory_used / self.GPU_memory_total

    def GPU_update(self):
        GPUs = getGPUs()
        self.GPU_memory_used = 0
        self.GPU_memory_free = 0
        self.GPU_memory_load = 0
        if len(GPUs) > 0:
            for g in GPUs:
                self.GPU_memory_used += g.memoryUsed
                self.GPU_memory_free += g.memoryFree
            self.GPU_memory_load = self.GPU_memory_used / self.GPU_memory_total

    def _explore_memory(self, unit=Mem_MBytes):
        """
        Note: everything is in MBytes
        """
        self.MEMORY_unit_value = unit
        if unit == Mem_KBytes:
            self.MEMORY_unit = "kBytes"
        elif unit == Mem_MBytes:
            self.MEMORY_unit = "MBytes"
        else:
            self.MEMORY_unit = "GBytes"
        self.MEMORY_size = psutil.virtual_memory().total / self.MEMORY_unit_value
        self.MEMORY_available = (
            psutil.virtual_memory().available / self.MEMORY_unit_value
        )
        self.MEMORY_percent = psutil.virtual_memory().percent
        self.MEMORY_used = psutil.virtual_memory().used / self.MEMORY_unit_value
        self.MEMORY_free = psutil.virtual_memory().free / self.MEMORY_unit_value
        self.MEMORY_active = psutil.virtual_memory().active / self.MEMORY_unit_value
        self.MEMORY_inactive = psutil.virtual_memory().inactive / self.MEMORY_unit_value
        if "linux" not in self.OS_name:
            self.MEMORY_wired = psutil.virtual_memory().wired / self.MEMORY_unit_value
        else:
            self.MEMORY_wired = None

    def memory_update(self):
        current_MEMORY_available = (
            psutil.virtual_memory().available / self.MEMORY_unit_value
        )
        current_MEMORY_percent = psutil.virtual_memory().percent
        current_MEMORY_used = psutil.virtual_memory().used / self.MEMORY_unit_value
        current_MEMORY_free = psutil.virtual_memory().free / self.MEMORY_unit_value
        current_MEMORY_active = psutil.virtual_memory().active / self.MEMORY_unit_value
        current_MEMORY_inactive = (
            psutil.virtual_memory().inactive / self.MEMORY_unit_value
        )
        # compute deltas
        delta_available = current_MEMORY_available - self.MEMORY_available
        delta_percent = current_MEMORY_percent - self.MEMORY_percent
        delta_used = current_MEMORY_used - self.MEMORY_used
        delta_free = current_MEMORY_free - self.MEMORY_free
        delta_active = current_MEMORY_active - self.MEMORY_active
        delta_inactive = current_MEMORY_inactive - self.MEMORY_inactive
        # store new values
        self.MEMORY_available = current_MEMORY_available
        self.MEMORY_percent = current_MEMORY_percent
        self.MEMORY_used = current_MEMORY_used
        self.MEMORY_free = current_MEMORY_free
        self.MEMORY_active = current_MEMORY_active
        self.MEMORY_inactive = current_MEMORY_inactive
        return (
            delta_available,
            delta_percent,
            delta_used,
            delta_free,
            delta_active,
            delta_inactive,
        )

    def update(self):
        self.memory_update()
        if self.GPU_ncores > 0:
            self.GPU_update()

    def get_report(self):
        print(self)

    ### Sanity checks ###
    def sanity_check(self):
        # python checks
        if self.Python_version_major < NRV_PYTHON_VERSION["major"]:
            print("Warning: Python 2 not supported")
        elif self.Python_version_minor < NRV_PYTHON_VERSION["minor"]:
            display_str = (
                "Warning: Python version detected: "
                + str(self.Python_version_major)
                + "."
                + str(self.Python_version_minor)
                + "."
                + str(self.Python_version_micro)
                + "\n minimal python version is "
                + str(NRV_PYTHON_VERSION["major"])
                + "."
                + str(NRV_PYTHON_VERSION["minor"])
                + "."
                + str(NRV_PYTHON_VERSION["patch"])
            )
            print(display_str)



"""
NRV-:class:`NRV_singleton` handling.
"""

# from .log_interface import rise_warning
from multiprocessing import Lock

class NRV_singleton(type):
    """
    Should be used as metaclass to define singleton classes

    """
    _instances = {}

    # _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        # with cls._lock:
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]




"""Electrical Impedance Tomography - eit: models for neural impedance imaging

eit provides high level models for imaging techniques based on electrical
impedance sensing for neural activity reconstruction. Tissue properties are
computed taking into account the temporal and frequencial evolution of neural
membranes conductivity. Models are based on both Finite Differences (NEURON)
computations and FEM models.


.. warning::
  The eit sb-package is currently under construction and scientific validation.
  The code can change fast, results not guaranteed, and developpers do not
  ensure backwards compatibility.

"""

from ._eit_forward import eit_forward
from ._eit_inverse import eit_inverse
from ._eit_utils import crop_fascicle, crop_nerve
from ._protocol import protocol, pyeit_protocol

submodules = []

classes = [
    "eit_forward",
    "eit_inverse",
    "protocol",
    "pyeit_protocol",
]

functions = [
    "crop_fascicle",
    "crop_nerve",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



"""
NRV-:class:`.material` handling.
"""

import faulthandler
import os
import numpy as np
from scipy.constants import pi, epsilon_0

from ..backend._log_interface import rise_warning
from ..backend._NRV_Class import NRV_class
from ..backend._parameters import parameters

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

# get the built-in material librairy
dir_path = parameters.nrv_path + "/_misc"
material_library = os.listdir(dir_path + "/materials/")

###############
## Functions ##
###############


def is_mat(mat):
    """
    check if an object is a material, return True if yes, else False

    Parameters
    ----------
    mat : object
        object to test

    Returns
    -------
    bool
        True it the type is a material object
    """
    return isinstance(mat, material)


def get_mat_file_as_dict(fname):
    """
    Open .mat material librairy file and return all lines as a dictionnary

    Returns
    -------
    d   : dictionnary
        physical properties of the material
    """
    d = {}
    with open(fname, "r") as f:
        for line in f:
            (key, value) = line.split()
            d[key] = value
    return d


def load_material(f_material):
    """
    Load a material by its name. if the name is in the NRV2 material librairy, the extention is
    automatically added.

    Parameters
    ----------
    f_material  : str
        either material name if the material is in the NRV2 Librairy or path to the corresponding\
        .mat material file
    """
    # load from librairy or from file
    f_in_librairy = str(f_material) + ".mat"
    if f_in_librairy in material_library:
        mat_file = get_mat_file_as_dict(dir_path + "/materials/" + f_in_librairy)
    else:
        mat_file = get_mat_file_as_dict(f_material)
    # creat material instance
    mat_obj = material()
    if "name" in mat_file:
        mat_obj.set_name(mat_file["name"])
    if "source" in mat_file:
        mat_obj.set_source(mat_file["source"])
    if "sigma_xx" in mat_file:
        mat_obj.set_anisotropic_conductivity(
            mat_file["sigma_xx"], mat_file["sigma_yy"], mat_file["sigma_zz"]
        )
    elif "sigma" in mat_file:
        mat_obj.set_isotropic_conductivity(mat_file["sigma"])
    else:
        rise_warning(
            "loading a material with 0 conductivity, \
            this may induce further division by 0"
        )
    if "epsilon_r" in mat_file:
        mat_obj.set_permitivity(mat_file["epsilon_r"])
    return mat_obj


def compute_effective_conductivity(sigma: float, epsilon: float, freq: float) -> float:
    r"""
    return the effective conductivity of the material.
    Two cases are psooible:

        - purely conductive material (when both `epsilon` or `freq` are set)

        .. math:: \sigma_{eff} = \sigma

        - dielectric material at a fixe frequency:

        .. math:: \sigma_{eff} = |\sigma+2j\pi f\epsilon_0\epsilon_r|

    Parameters
    ----------
    sigma   : float
        conductivity
    """
    if epsilon is None or freq is None:
        return sigma
    return abs(sigma + 2j * pi * freq * epsilon_0 * epsilon)


####################
## material class ##
####################
class material(NRV_class):
    """
    a class for material, where all the physical properties constants are stored.

    by default materials in NRV are considerated as purely conductive. Yet,
    """

    def __init__(self):
        """
        material instantiation
        """
        super().__init__()
        self.name = ""
        self.source = ""

        # Conduction properties
        self.isotrop_cond = True
        self._sigma = 0
        self._sigma_xx = 0
        self._sigma_yy = 0
        self._sigma_zz = 0
        self._sigma_func = None

        # permitivity property
        self.epsilon = None
        self.freq = None

    @property
    def sigma(self):
        """
        get the conductivity of the material
        """
        if not self.is_isotropic():
            return np.array([self.sigma_xx, self.sigma_yy, self.sigma_zz])
        return compute_effective_conductivity(
            sigma=self._sigma, epsilon=self.epsilon, freq=self.freq
        )

    @property
    def sigma_xx(self) -> float:
        """
        get the conductivity of the material along ox
        """
        if self.isotrop_cond:
            rise_warning("Isotropic conductor sigma_xx is sigma")
            return self.sigma
        return compute_effective_conductivity(
            sigma=self._sigma_xx, epsilon=self.epsilon, freq=self.freq
        )

    @property
    def sigma_yy(self) -> float:
        """
        get the conductivity of the material along oy
        """
        if self.isotrop_cond:
            rise_warning("Isotropic conductor sigma_yy is sigma")
            return self.sigma
        return compute_effective_conductivity(
            sigma=self._sigma_yy, epsilon=self.epsilon, freq=self.freq
        )

    @property
    def sigma_zz(self) -> float:
        """
        get the conductivity of the material along oz
        """
        if self.isotrop_cond:
            rise_warning("Isotropic conductor sigma_zz is sigma")
            return self.sigma
        return compute_effective_conductivity(
            sigma=self._sigma_zz, epsilon=self.epsilon, freq=self.freq
        )

    def save_material(self, save=False, fname="material.json"):
        rise_warning("save_material is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_material(self, data="material.json"):
        rise_warning("load_material is a deprecated method use load")
        self.load(data=data)

    def set_name(self, name: str) -> None:
        """
        set a name to a material

        Parameters
        ----------
        name :  str
            name of the material
        """
        self.name = name

    def set_source(self, source: str) -> None:
        """
        set a source to a material. Note that this source is a string without spaces

        Parameters
        ----------
        source  : str
            scientific reference to the value, for clarity only
        """
        self.source = source

    def set_isotropic_conductivity(self, sigma: float|complex) -> None:
        """
        set the conductivity for an isotropic material

        Parameters
        ----------
        sigma   : float
            conductivity in S/m
        """
        self.isotrop_cond = True
        if isinstance(sigma, complex):
            self._sigma = sigma
        else:
            self._sigma = float(sigma)


    def set_anisotropic_conductivity(
        self, sigma_xx: float, sigma_yy: float, sigma_zz: float
    ) -> None:
        """
        set the conductivity tensor for an anisotropic material

        Parameters
        ----------
        sigma_xx    : float
            conductivity along the x axis, in S/m
        sigma_yy    : float
            conductivity along the y axis, in S/m
        sigma_zz    : float
            conductivity along the z axis, in S/m
        """
        self.isotrop_cond = False
        self._sigma_xx = float(sigma_xx)
        self._sigma_yy = float(sigma_yy)
        self._sigma_zz = float(sigma_zz)

    def is_isotropic(self) -> bool:
        """
        check that the material is isotropic or not, return true if isotropic, else false

        Returns
        -------
        bool    :
            True if the per
        """
        return self.isotrop_cond

    ## Permitivity relatied methods
    def set_permitivity(self, epsilon: float) -> None:
        """
        set the relative permitivity of the material.

        Note
        ----
        The relative permetivity is used to compute an apparent conductivity at a frequency
        when ``freq``-attribute is set.

        Parameters
        ----------
        epsilon   : float
            relative permetivity in F/m
        """
        self.epsilon = float(epsilon)

    def set_frequency(self, freq: float) -> None:
        """
        set the frequency of the electric field in the material

        Parameters
        ----------
        freq   : float
            frequency of the fields in to set in kHz
        """
        self.freq = float(freq)

    def clear_frequency(self) -> None:
        """
        set the woking frequency to None

        Parameters
        ----------
        freq   : float
            frequency of the fields in to set in kHz
        """
        self.freq = None

    def is_permitive(self):
        """
        check if the material permitivity and the electric field frequency are set
        """
        return self.epsilon is not None and self.freq is not None



""" FIELD Models - fmod: handles extracellular field models.

fmod hosts the code to compute extracellular electrical fields and quantities.
Such quantity can result from both:


* electrical stimulation, handled by injecting stimulation current waveforms
  on electrodes. Associated computations can be performed using analitical 
  approach (fast but relying on strong hypotheses), or using Finite Element 
  models,
* the activity of the cells. In this second case, computations are for the 
  moment only performed with an analitical approach.


Finite Elements solver can be chosen between COMSOL (requieres extra license, 
this is not the recommended choice and is maintained only for comparison with 
existing results in the litterature) and FenicsX. This last solution is fully open
source and should be preferred. In this case, geometries are meshed using GMSH. All FEM
computations are handled by a subpackage called ``FEM`` (seel below).

.. note::
  for scientific details of how nmod works and an overview of the general
  implementation, refer to the 'Scientific foundations' section of the
  documentation.

"""


from ._electrodes import (
    is_FEM_electrode,
    is_CUFF_electrode,
    is_LIFE_electrode,
    is_analytical_electrode,
    load_any_electrode,
    check_electrodes_overlap,
    electrode,
    point_source_electrode,
    FEM_electrode,
    LIFE_electrode,
    CUFF_electrode,
    CUFF_MP_electrode,
    
)
from ._extracellular import (
    is_extra_stim,
    is_analytical_extra_stim,
    is_FEM_extra_stim,
    load_any_extracel_context,
    extracellular_context,
    stimulation,
    FEM_stimulation,
)
from ._materials import (
    is_mat,
    get_mat_file_as_dict,
    load_material,
    compute_effective_conductivity,
    material,
)
from ._recording import (
    is_recording_point,
    is_recorder,
    NodeD_interpol,
    recording_point,
    recorder,
)

from . import FEM
submodules = [
    "FEM",
]

classes = [
    "material",
    "electrode",
    "point_source_electrode",
    "FEM_electrode",
    "LIFE_electrode",
    "CUFF_electrode",
    "CUFF_MP_electrode",
    "extracellular_context",
    "stimulation",
    "FEM_stimulation",
    "recording_point",
    "recorder",
]

functions = [
    "is_FEM_electrode",
    "is_CUFF_electrode",
    "is_LIFE_electrode",
    "is_analytical_electrode",
    "is_extra_stim",
    "is_analytical_extra_stim",
    "is_FEM_extra_stim",
    "is_mat",
    "is_recording_point",
    "is_recorder",
    "load_any_electrode",
    "load_any_extracel_context",
    "load_material",
    "check_electrodes_overlap",
    "get_mat_file_as_dict",
    "compute_effective_conductivity",
    "NodeD_interpol",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



"""
NRV-:class:`.FEM_model` handling.
"""

from abc import abstractmethod

from ...backend._log_interface import pass_info, rise_warning
from ...backend._NRV_Class import NRV_class

###############
## Constants ##
###############
fem_verbose = True


###################
## Model classes ##
###################
class FEM_model(NRV_class):
    """
    A generic class for Finite Element models
    """

    @abstractmethod
    def __init__(self, n_proc=None):
        """
        Creates a instance of FEM_model

        Parameters
        ----------
        n_proc   : int
            number of cores for computation. If None is specified, this number is taken from the NRV2.ini configuration file. Byu default set to None
        """
        super().__init__()
        self.n_proc = n_proc
        self.is_multi_proc = False
        self.type = "FEM"

        # Timmers
        self.meshing_timer = 0
        self.setup_timer = 0
        self.solving_timer = 0
        self.access_res_timer = 0

        # Flags
        self.is_meshed = False
        self.is_computed = False

    def get_timers(self, verbose=False):
        """
        Returns the timers of each step of the FEM computation

        Parameters
        ----------
        verbose     : bool
            if true, timer are also passed to the log and the terminal (if verbosity level > 2)

        Returns
        -------
        meshing_timer   : float
            time spent to mesh the FEM problem, in s
        setup_timer   : float
            time spent to mesh the setup (i.e. generating variab connecting to the server), in s
        solving_timer   : float
            time spent to mesh the FEM problem, in s
        total_timer   : float
            time spent to mesh the FEM problem, in s
        """
        pass_info("mesh done in " + str(self.meshing_timer) + " s")
        pass_info("simulation setup in " + str(self.setup_timer) + " s")
        pass_info("simulation solved in " + str(self.solving_timer) + " s")
        pass_info("Time spent to access results " + str(self.access_res_timer) + " s")
        total_timer = (
            self.meshing_timer
            + self.setup_timer
            + self.solving_timer
            + self.access_res_timer
        )
        pass_info("total duration " + str(total_timer) + " s")
        return self.meshing_timer, self.setup_timer, self.solving_timer, total_timer





"""
NRV-:class:`.unmyelinated_results` handling.
"""

import numpy as np
import matplotlib.pyplot as plt
from ._axons_results import axon_results


class unmyelinated_results(axon_results):
    """ """

    def __init__(self, context=None):
        super().__init__(context)

    def generate_axon(self):
        if "unmyelinated" not in globals():
            from .._unmyelinated import unmyelinated
        return unmyelinated(**self)

    def plot_x_t(
        self,
        axes: plt.Axes,
        key: str = "V_mem",
        color: str = "k",
        n_lines: int = 20,
        n_jumped_lines: int|None = None,
        switch_axes=False,
        norm = None,
        **kwgs
    ) -> None:
        if n_jumped_lines is not None:
            x_index = np.arange(len(self.x_rec))
            x_index = x_index[x_index%n_jumped_lines==0]
        else:
            x_index = np.int32(np.linspace(0, len(self.x_rec) - 1, n_lines))
        x_pos = self.x_rec[x_index]
        super().plot_x_t(axes=axes, x_pos=x_pos,x_index=x_index, key=key, color=color, switch_axes=switch_axes, norm=norm, **kwgs)



""" NEURON Models - nmod: handles neural fiber models using the NEURON software.

nmod contains all the code to describe the axonal fibers using .mod mechanisms
with NEURON. This subpackage contains all primitives to describe:

  - unmyelinated axons with different models of nonlinear conductances,
  - myelinated axons with different models of nonlinear conductances,
  - fascicles consisting in bundles of (potentiallymixed ) axons,
  - nerves containing one or more fascicles.

nmod also contains tools to automatically generate fascicles and nerves,
with fiber repartition mimicking realistic anatomical observations.

Classes describing fibers, fascicles and nerves are simulable, meaning they
all have a `simulate` method that triggers the simulation. Only the top level
structure of the simulated scenario has to be simulated by the user, nrv takes
care of handling the simulation, potential combination with subpackages
(such as nmod or eit for instance). Parallel computing is also handled
internally.

.. SeeAlso::
   :doc:`Scientific details</scientific>`, :doc:`Simulables users guide</usersguide/simulables>`
"""

from ._axon_pop_generator import (
    load_stat,
    create_axon_population,
    fill_area_with_axons,
    axon_packer,
    expand_pop,
    remove_collision,
    remove_outlier_axons,
    get_circular_contour,
    plot_population,
    save_axon_population,
    load_axon_population,
)
from ._axons import axon
from ._unmyelinated import unmyelinated
from ._myelinated import myelinated
from ._fascicles import fascicle
from ._nerve import nerve
from . import results

submodules = ["results"]

classes = [
    "axon",
    "unmyelinated",
    "myelinated",
    "fascicle",
    "nerve"]

functions = [
    "create_axon_population",
    "plot_population",
    "save_axon_population",
    "load_axon_population",
    "fill_area_with_axons",
    "axon_packer",
    "expand_pop",
    "remove_collision",
    "remove_outlier_axons",
    "get_circular_contour",
    "load_stat",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, PillowWriter, FFMpegWriter
from ._OptimFunctions import interpolate_Npts
from .._CostFunctions import cost_function
import numpy as np

from ...backend._NRV_Results import NRV_results
from ...backend._log_interface import pass_info, rise_warning


class optim_results(NRV_results):
    def __init__(self, context=None):
        super().__init__({"optimization_parameters": context})

    def load(self, data, blacklist=[], **kwargs):
        super().load(data, blacklist, **kwargs)
        self["optimization_parameters"] = self["optimization_parameters"].save(
            save=False
        )

    ############################
    ##   Processing methods   ##
    ############################

    ## PSO relative methods
    def is_stabilized(self, part, it, threshold=0.1):
        velocity = self["velocity" + str(part)][it]
        for i in velocity:
            if abs(i) >= threshold:
                return False
        return True

    def stabilization_it(self, parts=None, nit=None, threshold=None):
        if threshold == None:
            threshold = max(max(self["velocity1"])) / 100
            pass_info("No threshold in parameters, set to Vpart1max/100 = ", threshold)

        # One particle
        if type(parts) == int:
            if nit is None:
                nit = self.nit
            for i in range(nit - 1, 0, -1):
                if not self.is_stabilized(parts, i, threshold):
                    return i
            return None

        # A particle list
        elif np.iterable(parts):
            list_it = []
            for i in parts:
                list_it += [
                    self.stabilization_it(parts=int(i), nit=nit, threshold=threshold)
                ]
            return np.array(list_it)

        # The whole swram
        else:
            swarm = 1 + np.arange(self["optimization_parameters"]["n_particles"])
            return self.stabilization_it(swarm, nit=nit, threshold=threshold)

    def add_filter(self, part_filter):
        n_particles = self["optimization_parameters"]["n_particles"]
        nit = self.nit
        resultsf = self
        for i in range(n_particles):
            resultsf["position" + str(i + 1) + " filtered"] = [[] for j in range(nit)]
            for j in range(nit):
                part = resultsf["position" + str(i + 1)][j]
                filtered_part = part_filter(part)
                resultsf["position" + str(i + 1) + " filtered"][j] = filtered_part
        return resultsf

    def findbestpart(self, decimals=10, verbose=False, lim_it=None):
        if decimals < -15:
            rise_warning("Best results not founded returning -1")
            return -1
        n_particles = self["optimization_parameters"]["n_particles"]
        nit = self.nit
        bestpos = self["best_position"]
        ibestpart = 0

        if lim_it is None:
            lim_it = nit

        for j in range(lim_it):
            for i in range(1, 1 + n_particles):
                pos = self["position" + str(i)]
                if all(np.around(pos[j], decimals) == np.around(bestpos, decimals)):
                    if verbose:
                        print("it=", j)
                    return i

        if verbose:
            pass_info("not found with decimals =", decimals)
        return self.findbestpart(decimals - 1, verbose=verbose, lim_it=lim_it)

    def compute_best_pos(self, cost_function: cost_function, **kwrgs):
        return cost_function.get_sim_results(self.x)

    ############################
    ##    plotting methods    ##
    ############################
    def plot_cost_history(
        self,
        ax: plt.axes,
        nitstop: int = -1,
        xlog: bool = False,
        ylog: bool = False,
        **ax_kwargs
    ):
        cost = self["cost_history"]
        ax.plot(cost[0:nitstop], **ax_kwargs)
        if xlog:
            ax.set_xscale("log")
        if ylog:
            ax.set_yscale("log")



"""
NRV-:class:`.fascicle` handling.
"""

import faulthandler
import os

import matplotlib.pyplot as plt
import numpy as np
import multiprocessing as mp
from rich import progress

from ..backend._parameters import parameters
from ..backend._file_handler import *
from ..backend._log_interface import pass_info, rise_warning
from ..backend._NRV_Simulable import NRV_simulable
from ..fmod._extracellular import *
from ..fmod._recording import *
from ..ui._axon_postprocessing import *
from ..backend._inouts import check_function_kwargs
from ._axons import *
from ._axon_pop_generator import *
from ._myelinated import *
from ._unmyelinated import *
from .results._fascicles_results import fascicle_results
from .results._axons_results import axon_results



# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()


builtin_postproc_functions = {
    "default": default_PP,
    "default_PP": default_PP,
    "rmv_keys": rmv_keys,
    "is_recruited": is_recruited,
    "is_blocked": is_blocked,
    "is_onset": is_blocked,
    "sample_keys": sample_keys,
    "vmem_plot": vmem_plot,
    "raster_plot": raster_plot,
}

deprecated_builtin_postproc_functions = {
    "is_excited": "is_recruited",
    "save_gmem": "sample_g_mem",
}

#####################
## Fasscicle class ##
#####################


def is_fascicle(object):
    """
    check if an object is a fascicle, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a fascicle object
    """
    return isinstance(object, fascicle)


class fascicle(NRV_simulable):
    r"""
    Class for Fascicle, defined as a group of axons near one to the other in the same Perineurium Sheath. All axons are independant of each other, no ephaptic coupling.


    See Also
    --------
    :doc:`Simulables users guide</usersguide/simulables>`, :class:`Nerve-class<._nerve.nerve>`, :class:`Axon-class<._axons.axon>`


    .. rubric:: Customizable Attributes:

    .. list-table::
       :widths: 10 10 10 70
       :header-rows: 1

       * - Attributes
         - Type
         - Default
         - Description
       * - ``ID``
         - ``int``
         - 0
         - Identification number of the fascicle.
       *
         - ``L``
         - ``float``
         - None
         - Length of the fascicle.
       *
         - ``D``
         - ``float``
         - None
         - Diameter of the fascicle.
       *
         - ``y_grav_center``
         - ``float``
         - 0
         - y-position of the fascicle center.
       *
         - ``z_grav_center``
         - ``float``
         - 0
         - z-position of the fascicle center.
       *
         - ``postproc_label``
         - ``str``
         - None
         - Label of the axon postprocessing funtion, used for the buildin postproc functions.
       *
         - ``postproc_function``
         - ``function``
         - None
         - Axon postprocessing funtion, used for the custom postproc functions.
       *
         - ``postproc_script``
         - ``str`` | ``function``
         - None
         - Either postprocessing funtion or postprocessing funtion label, automatically set depending on the type
       *
         - ``postproc_kwargs``
         - ``dict``
         - None
         - key arguments of the postporcessing function
       *
         - ``save_results``
         - ``bool``
         - False
         - If ``True``, fascicle configuration and all axon simulations results are saved in ``save_path`` directory.
       *
         - ``save_path``
         - ``str``
         - ""
         - Path of the directory where simulation results should be saved.
       *
         - ``return_parameters_only``
         - ``bool``
         - False
         - If ``True`` (and ``save_results`` also ``True``), only the parameters should be returned from the simulation.
       *
         - ``loaded_footprints``
         - ``bool``
         - False
         - If ``False``, the footprints already computed are favored over new footprint computation.
       *
         - ``verbose``
         - ``bool``
         - False
         - Plot or not.


    Note
    ----
    Customizable attributes can either be set using :meth:`.fascicle.set_parameters` or simply by reafecting the value of the attribute.

    Tip
    ---
    Additional simulation parameters can be changed using (:meth:`.fascicle.set_axons_parameters`, :meth:`.fascicle.change_stimulus_from_electrode`, ...).



    """

    def __init__(self, diameter=None, ID=0, **kwargs):
        """
        Instantation of a Fascicle
        """
        super().__init__(**kwargs)

        # to add to a fascicle/nerve common mother class

        #:str: path where the simulation results should be saved
        self.save_path:str = ""
        #:str: value: False: verbosity mainly for pbars. Tests more comment
        self.verbose = False 
        self.return_parameters_only = False
        self.loaded_footprints = False
        self.save_results = False
        self.result_folder_name:str = ""

        self.postproc_label = "default"
        self.postproc_function = None
        self.postproc_script = None
        self.postproc_kwargs = {}

        self.config_filename = ""
        self.ID = ID
        self.type = None
        self.L = None
        self.D = diameter
        # geometric properties
        self.y_grav_center = 0
        self.z_grav_center = 0

        # Update parameters with kwargs
        self.set_parameters(**kwargs)

        # axonal content
        self.axons_diameter = np.array([])
        self.axons_type = np.array([], dtype=int)
        self.axons_y = np.array([])
        self.axons_z = np.array([])
        self.NoR_relative_position = np.array([])
        # Axons objects default parameters
        self.unmyelinated_param = {
            "model": "Rattay_Aberham",
            "dt": 0.001,
            "Nrec": 0,
            "Nsec": 1,
            "Nseg_per_sec": 100,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "v_init": None,
            "T": None,
            "threshold": -40,
        }

        self.myelinated_param = {
            "model": "MRG",
            "dt": 0.001,
            "Nseg_per_sec": 3,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "rec": "nodes",
            "T": None,
            "threshold": -40,
        }
        self.set_axons_parameters(**kwargs)

        # extra-cellular stimulation
        self.extra_stim = None
        self.footprints = {}
        self.is_footprinted = False
        # intra-cellular stimulation
        self.N_intra = 0
        self.intra_stim_position = []
        self.intra_stim_t_start = []
        self.intra_stim_duration = []
        self.intra_stim_amplitude = []
        self.intra_stim_ON = []
        ## recording mechanism
        self.record = False
        self.recorder = None
        # Simulation status
        self.is_simulated = False

    ## save/load methods
    def save(
        self,
        fname="fascicle.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        save=True,
        _fasc_save=True,
        blacklist=[],
    ):
        """
        Save a fascicle in a json file

        Parameters
        ----------
        fname :             str
            name of the file to save the fascicle
        extracel_context:   bool
            if True, add the extracellular context to the saving
        intracel_context:   bool
            if True, add the intracellular context to the saving
        rec_context:        bool
            if True, add the recording context to the saving
        blacklist:          list[str]
            key to exclude from saving
        save:               bool
            if false only return the dictionary

        Returns
        -------
        key_dict:           dict
            Python dictionary containing all the fascicle information
        """
        bl = [i for i in blacklist]
        to_save = (save and _fasc_save)

        bl += ["postproc_function", "postproc_script"]
        if self.postproc_label not in builtin_postproc_functions:
            rise_warning(
                "custom axon postprocessing cannot be save. Be carefull to set the postproc_function again when reloading fascicle"
            )

        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += ["extra_stim", "footprints", "is_footprinted"]

        if not rec_context:
            bl += ["record", "recorder"]

        return super().save(fname=fname, save=to_save, blacklist=bl)

    def load(
        self,
        data,
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Load a fascicle configuration from a json file

        Parameters
        ----------
        fname           : str
            path to the json file describing a fascicle
        extracel_context: bool
            if True, load the extracellular context as well
        intracel_context: bool
            if True, load the intracellular context as well
        rec_context: bool
            if True, load the recording context as well
        blacklist       : list[str]
            key to exclude from loading
        """
        if type(data) == str:
            key_dic = json_load(data)
        else:
            key_dic = data

        bl = [i for i in blacklist]
        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += [
                "extra_stim",
                "footprints",
                "myelinated_nseg_per_sec",
                "unmyelinated_nseg",
                "is_footprinted",
            ]

        if not rec_context:
            bl += ["record", "recorder"]

        super().load(data=key_dic, blacklist=bl)
        if "unmyelinated_param" not in key_dic:
            rise_warning("Deprecated fascicle file")
            self.set_axons_parameters(key_dic)

    def save_fascicle_configuration(
        self, fname, extracel_context=False, intracel_context=False, rec_context=False
    ):
        rise_warning(
            "save_fascicle_configuration is a deprecated method use save instead"
        )
        self.save(
            fname=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    def load_fascicle_configuration(
        self, fname, extracel_context=False, intracel_context=False, rec_context=False
    ):
        rise_warning(
            "load_fascicle_configuration is a deprecated method use load instead"
        )
        self.load(
            data=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    ## Fascicle property method
    @property
    def n_ax(self):
        """
        Number of axons in the fascicle
        """
        return len(self.axons_diameter)

    @property
    def N(self):
        """
        :meta private:
        """
        rise_warning(
            "DeprecationWarning: ",
            "fascicle.N property is obsolete use fascicle.n_ax instead",
        )
        return self.n_ax

    @property
    def A(self):
        """
        Area of the fascicle
        """
        if self.D is None:
            return None
        return np.pi * (self.D / 2) ** 2

    def set_ID(self, ID):
        """
        set the ID of the fascicle

        Parameters
        ----------
        ID : int
            ID number of the fascicle
        """
        self.ID = ID

    def define_length(self, L):
        """
        set the length over the x axis of the fascicle

        Parameters
        ----------
        L   : float
            length of the fascicle in um
        """
        if self.extra_stim is not None or self.N_intra > 0:
            rise_warning(
                "Modifying length of a fascicle with extra or intra cellular context can lead to error"
            )
        self.L = L
        self.unmyelinated_param["Nseg_per_sec"] == self.L // 25

    ## generate stereotypic Fascicle
    def define_circular_contour(self, D, y_c=None, z_c=None):
        """
        Define a circular countour to the fascicle

        Parameters
        ----------
        D           : float
            diameter of the circular fascicle contour, in um
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        """
        self.type = "Circular"
        self.D = D
        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c

    def get_circular_contour(self):
        """
        Returns the properties of the fascicle contour considered as a circle (y and z center and diameter)

        Returns
        -------
        D : float
            diameter of the contour, in um. Set to 0 if not applicable
        y : float
            y position of the contour center, in um
        z : float
            z position of the contour center, in um
        """
        y = self.y_grav_center
        z = self.z_grav_center
        D = self.D
        return D, y, z

    def fit_circular_contour(self, y_c=None, z_c=None, delta=0.1, round_dgt=None):
        """
        Define a circular countour to the fascicle

        Parameters
        ----------
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        delta       : float
            distance between farest axon and contour, in um
        """
        rise_warning(
            "fit_circular_contour method usage is not recommended anymore and will be removed in future release."
        )
        pass_info("Define fascicle size/shape at object creation instead.")
        N_axons = len(self.axons_diameter)
        D = 2 * delta

        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        if N_axons == 0:
            pass_info("No axon to fit fascicul diameter set to " + str(D) + "um")
        else:
            dist_axons = np.linalg.norm((self.axons_y, self.axons_z), axis=0) + (self.axons_diameter/2)
            D = 2 * (dist_axons.max() + delta)
        if round is not None:
            D = round(D, round_dgt)
        self.define_circular_contour(D, y_c=None, z_c=None)

    def define_ellipsoid_contour(self, a, b, y_c=0, z_c=0, rotate=0):
        """
        Define ellipsoidal contour
        """
        pass

    ## generate Fascicle from histology
    def import_contour(self, smthing_else):
        """
        Define contour from a file
        """
        pass

    ## fill fascicle methods
    def fill(
        self,
        percent_unmyel=0.7,
        FVF=0.55,
        M_stat="Schellens_1",
        U_stat="Ochoa_U",
        ppop_fname=None,
        pop_fname=None,
    ):
        """
        Fill a geometricaly defined contour with axons

        Parameters
        ----------
        parallel        : bool
            if True, the generation process (quite long) is split over multiples cores, if False everything is perfrmed by the master.
        percent_unmyel  : float
            ratio of unmyelinated axons in the population. Should be between 0 and 1.
        FVF             : float
            Fiber Volume Fraction estimated for the area. By default set to 0.55
        M_stat          : str
            name of the statistic in the librairy or path to a new librairy in csv for myelinated diameters repartition
        U_stat          : str
            name of the statistic in the librairy or path to a new librairy in csv for unmyelinated diameters repartition
        ppop_fname      : str
            optional, if specified, name file to store the placed population generated
        pop_fname       : str
            optional, if specified, name file to store the population generated
        """
        rise_warning(
            "fill method usage is not recommended anymore and will be removed in future release."
        )
        pass_info("Use axon_pop_generator methods instead.")
        #### AXON GENERATION: parallelization if resquested ####
        Area_to_fill = 0
        # Note: generate a bit too much axons just in case
        if self.type == "Circular":
            Area_to_fill = np.pi * (self.D / 2 + 28) ** 2
        (
            axons_diameter,
            axons_type,
            M_diam_list,
            U_diam_list,
        ) = fill_area_with_axons(
            Area_to_fill,
            percent_unmyel=percent_unmyel,
            FVF=FVF,
            M_stat=M_stat,
            U_stat=U_stat,
        )
        #### AXON PACKING: never parallel
        N = len(axons_diameter)
        pass_info("\n ... " + str(N) + " axons generated")
        if pop_fname is not None:
            save_axon_population(
                pop_fname, axons_diameter, axons_type, comment=None
            )
        (
            axons_y,
            axons_z,
        ) = axon_packer(
            axons_diameter,
            delta=0.1,
            y_gc=self.y_grav_center,
            z_gc=self.z_grav_center,
            n_iter=20000,
        )

        N = len(axons_diameter)
        # check if axons are inside the fascicle
        inside_axons = (
            np.power(axons_y - self.y_grav_center, 2)
            + np.power(axons_z - self.z_grav_center, 2)
            - np.power(np.ones(N) * (self.D / 2) - axons_diameter / 2, 2)
        )
        axons_to_keep = np.argwhere(inside_axons < 0)
        axons_diameter = axons_diameter[axons_to_keep]
        axons_type = axons_type[axons_to_keep]
        axons_y = axons_y[axons_to_keep]
        axons_z = axons_z[axons_to_keep]
        N = len(axons_diameter)
        # save the good population
        if ppop_fname is not None:
            save_axon_population(
                ppop_fname,
                self.axons_diameter,
                self.axons_type,
                self.axons_y,
                self.axons_z,
                comment=None,
            )


    def fill_with_population(
        self,
        axons_diameter: np.array,
        axons_type: np.array,
        delta: float = 1,
        y_axons: np.array = None,
        z_axons: np.array = None,
        fit_to_size: bool = False,
        n_iter: int = 20_000,
    ) -> None:
        """
        Fill the fascicle with an axon population

        Parameters
        ----------
        axons_diameters     : np.array
            Array  containing all the diameters of the axon population
        axons_type          : np.array
            Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
        delta               : float
            axon-to-axon and axon to border minimal distance
        y_axons             : np.array
            y coordinate of the axon population, in um
        z_axons             : np.array
            z coordinate of the axons population, in um
        fit_to_size         : bool
            if true, the axon population is extended to fit within fascicle size, if not the population is kept as is
        n_iter              : int
            number of interation for the packing algorithm if the y-x axon coordinates are not specified
        WARNING: If y_axons and z_axons are not specified then axon-packing algorithm is called within the method.
        """
        N = len(axons_diameter)
        if (y_axons is None) and (z_axons is None):
            y_axons, z_axons = axon_packer(
                axons_diameter, delta=delta, n_iter=n_iter
            )
        if fit_to_size:
            if self.D is not None:
                d_pop = get_circular_contour(
                    axons_diameter, y_axons, z_axons, delta
                )
                if (d_pop) < self.D:
                    exp_factor = 0.99 * (self.D / d_pop)
                    y_axons, z_axons = expand_pop(y_axons, z_axons, exp_factor)
            else:
                rise_warning(
                    "Can't fit population to size, fascicle diameter is not defined."
                )

        axons_diameter, y_axons, z_axons, axons_type = remove_collision(
            axons_diameter, y_axons, z_axons, axons_type
        )
        if self.D is not None:
            axons_diameter, y_axons, z_axons, axons_type = remove_outlier_axons(
                axons_diameter, y_axons, z_axons, axons_type, self.D - delta
            )
        
        self.axons_diameter = axons_diameter.flatten()
        self.axons_type = axons_type.flatten()
        self.axons_y = y_axons.flatten()
        self.axons_z = z_axons.flatten()
        self.translate_axons(self.y_grav_center, self.z_grav_center)


    def fit_population_to_size(self, delta: float = 1):
        """
        Fit the axon positions to the size of the fascicle

        Parameters
        ----------
        delta   : float
            minimum axon to fascicle border distance, in um
        """
        self.translate_axons(-self.y_grav_center, -self.z_grav_center)
        d_pop = get_circular_contour(
            self.axons_diameter, self.axons_y, self.axons_z, delta
        )
        if (d_pop) < self.D:
            exp_factor = 0.99 * (self.D / d_pop)
            self.axons_y, self.axons_z = expand_pop(
                self.axons_y, self.axons_z, exp_factor
            )
        self.translate_axons(self.y_grav_center, self.z_grav_center)

    ## Move methods
    def translate_axons(self, y, z):
        """
        Move axons only in a fascicle by group translation

        Parameters
        ----------
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        self.axons_y += y
        self.axons_z += z

    def translate_fascicle(self, y, z):
        """
        Translate a complete fascicle

        Parameters
        ----------
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        self.y_grav_center += y
        self.z_grav_center += z
        self.translate_axons(y, z)
        if self.extra_stim is not None:
            self.extra_stim.translate(y=y, z=z)
        if self.recorder is not None:
            self.recorder.translate(y=y, z=z)

    def rotate_axons(self, theta, y_c=0, z_c=0):
        """
        Move axons only in a fascicle by group rotation

        Parameters
        ----------
        theta   : float
            angular value of the translation, in rad
        y_c     : float
            y axis value of the rotation center in um, by default set to 0
        z_c     : float
            z axis value for the rotation center in um, by default set to 0
        """
        self.axons_y = (
            np.cos(theta) * (self.axons_y - y_c) - np.sin(theta) * (self.axons_z - z_c)
        ) + y_c
        self.axons_z = (
            np.sin(theta) * (self.axons_y - y_c) + np.cos(theta) * (self.axons_z - z_c)
        ) + z_c

    def rotate_fascicle(self, theta, y_c=0, z_c=0):
        """
        Rotate a complete fascicle

        Parameters
        ----------
        theta   : float
            angular value of the translation, in rad
        y_c     : float
            y axis value of the rotation center in um, by default set to 0
        z_c     : float
            z axis value for the rotation center in um, by default set to 0
        """
        self.y_grav_center = (
            np.cos(theta) * (self.y_grav_center - y_c)
            - np.sin(theta) * (self.z_grav_center - z_c)
        ) + y_c
        self.z_grav_center = (
            np.sin(theta) * (self.y_grav_center - y_c)
            + np.cos(theta) * (self.z_grav_center - z_c)
        ) + z_c
        self.rotate_axons(theta, y_c=y_c, z_c=z_c)

    ## Axons related method
    def set_axons_parameters(
        self, unmyelinated_only=False, myelinated_only=False, **kwargs
    ):
        """
        set parameters of axons in the fascicle

        Parameters
        ----------
        unmyelinated_only:      bool
            if true modification only done for unmyelinated axons parameters, by default False
        myelinated_only:        bool
            if true modification only done for myelinated axons parameters, by default False
        kwargs:
            parameters to modify (see myelinated and unmyelinated)
        """
        ## Standard keys
        for key in kwargs:
            if "model" in key:
                if not myelinated_only and kwargs[key] in unmyelinated_models:
                    self.unmyelinated_param["model"] = kwargs[key]
                elif not unmyelinated_only and kwargs[key] in myelinated_models:
                    self.myelinated_param["model"] = kwargs[key]
                else:
                    rise_warning(kwargs[key], " is not an implemented axon model")
            else:
                if not myelinated_only and key in self.unmyelinated_param:
                    self.unmyelinated_param[key] = kwargs[key]
                if not unmyelinated_only and key in self.myelinated_param:
                    self.myelinated_param[key] = kwargs[key]

        ## Specific keys
        if "unmyelinated_nseg" in kwargs:
            self.unmyelinated_param["Nseg_per_sec"] = kwargs["unmyelinated_nseg"]
        if "myelinated_nseg_per_sec" in kwargs:
            self.myelinated_param["Nseg_per_sec"] = kwargs["myelinated_nseg_per_sec"]

    def get_axons_parameters(self, unmyelinated_only=False, myelinated_only=False):
        """
        get parameters of axons in the fascicle

        Parameters
        ----------
        unmyelinated_only:      bool
            modification will only
        myelinated_only:        bool
            modification will only
        """
        if unmyelinated_only:
            return self.unmyelinated_param
        if myelinated_only:
            return self.myelinated_param
        return self.unmyelinated_param, self.myelinated_param

    def __generate_axon(self,k:int)->axon:
        """
        Internal use only: generate fascicle't kth axon

        Parameters
        ----------
        k : int
            _description_
        """
        if self.axons_type[k] == 0:
            axon = unmyelinated(
                self.axons_y[k],
                self.axons_z[k],
                self.axons_diameter[k],
                self.L,
                ID=k,
                **self.unmyelinated_param,
            )
        else:
            self.myelinated_param["node_shift"] = self.NoR_relative_position[k]
            axon = myelinated(
                self.axons_y[k],
                self.axons_z[k],
                self.axons_diameter[k],
                self.L,
                ID=k,
                **self.myelinated_param,
            )
            self.myelinated_param.pop("node_shift")
        return axon

    def remove_unmyelinated_axons(self):
        """
        Remove all unmyelinated fibers from the fascicle
        """
        mask = self.axons_type.astype(bool)
        self.axons_diameter = self.axons_diameter[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]
        self.axons_type = self.axons_type[mask]
        if len(self.NoR_relative_position) != 0:
            self.NoR_relative_position = self.NoR_relative_position[mask]

    def remove_myelinated_axons(self):
        """
        Remove all myelinated fibers from the
        """
        mask = np.invert(self.axons_type.astype(bool))
        self.axons_diameter = self.axons_diameter[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]
        self.axons_type = self.axons_type[mask]
        if len(self.NoR_relative_position) != 0:
            # almost useless but here for coherence
            self.NoR_relative_position = self.NoR_relative_position[mask]

    def remove_axons_size_threshold(self, d, min=True):
        """
        Remove fibers with diameters below/above a threshold
        """
        if min:
            mask = self.axons_diameter >= d
        else:
            mask = self.axons_diameter <= d

        self.axons_diameter = self.axons_diameter[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]
        self.axons_type = self.axons_type[mask]
        if len(self.NoR_relative_position) != 0:
            # almost useless but here for coherence
            self.NoR_relative_position = self.NoR_relative_position[mask]

    ## Representation methods
    def plot(
        self,
        axes,
        contour_color="k",
        myel_color="b",
        unmyel_color="r",
        elec_color="gold",
        num=False,
    ):
        """
        plot the fascicle in the Y-Z plane (transverse section)

        Parameters
        ----------
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        contour_color   : str
            matplotlib color string applied to the contour. Black by default
        myel_color      : str
            matplotlib color string applied to the myelinated axons. Red by default
        unmyel_color    : str
            matplotlib color string applied to the myelinated axons. Blue by default
        num             : bool
            if True, the index of each axon is displayed on top of the circle
        """
        ## plot contour
        axes.add_patch(
            plt.Circle(
                (self.y_grav_center, self.z_grav_center),
                self.D / 2,
                color=contour_color,
                fill=False,
                linewidth=2,
            )
        )
        ## plot axons
        circles = []
        for k in range(self.n_ax):
            if self.axons_type[k] == 1:  # myelinated
                circles.append(
                    plt.Circle(
                        (self.axons_y[k], self.axons_z[k]),
                        self.axons_diameter[k] / 2,
                        color=myel_color,
                        fill=True,
                    )
                )
            else:
                circles.append(
                    plt.Circle(
                        (self.axons_y[k], self.axons_z[k]),
                        self.axons_diameter[k] / 2,
                        color=unmyel_color,
                        fill=True,
                    )
                )
        for circle in circles:
            axes.add_patch(circle)
        if self.extra_stim is not None:
            self.extra_stim.plot(axes=axes, color=elec_color, nerve_d=self.D)
        if num:
            for k in range(self.n_ax):
                axes.text(self.axons_y[k], self.axons_z[k], str(k))
        axes.set_xlim((-1.1 * self.D / 2, 1.1 * self.D / 2))
        axes.set_ylim((-1.1 * self.D / 2, 1.1 * self.D / 2))

    def plot_x(self, axes, myel_color="b", unmyel_color="r", Myelinated_model="MRG"):
        """
        plot the fascicle's axons along Xline (longitudinal)

        Parameters
        ----------
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        myel_color      : str
            matplotlib color string applied to the myelinated axons. Red by default
        unmyel_color    : str
            matplotlib color string applied to the myelinated axons. Blue by default
        Myelinated_model : str
            model use for myelinated axon (use to calulated node position)
        """
        if self.L is None or len(self.NoR_relative_position)>0:
            drange = [
                min(self.axons_diameter.flatten()),
                max(self.axons_diameter.flatten()),
            ]
            polysize = np.poly1d(np.polyfit(drange, [0.5, 5], 1))
            for k in range(self.n_ax):
                relative_pos = self.NoR_relative_position[k]
                d = self.axons_diameter[k]
                if self.axons_type.flatten()[k] == 0.0:
                    color = unmyel_color
                    size = polysize(d)
                    axes.plot([0, self.L], np.ones(2) + k - 1, color=color, lw=size)
                else:
                    color = myel_color
                    size = polysize(d)
                    axon = self.__generate_axon(k)
                    x_nodes = axon.x_nodes
                    node_number = len(x_nodes)
                    del axon
                    axes.plot([0, self.L], np.ones(2) + k - 1, color=color, lw=size)
                    axes.scatter(
                        x_nodes,
                        np.ones(node_number) + k - 1,
                        marker="x",
                        color=color,
                    )
            ## plot electrode(s) if existings
            if self.extra_stim is not None:
                for electrode in self.extra_stim.electrodes:
                    if not is_FEM_electrode(electrode):
                        axes.plot(
                            electrode.x * np.ones(2),
                            [0, self.n_ax - 1],
                            color="gold",
                        )
            axes.set_xlabel("x (um)")
            axes.set_ylabel("axon ID")
            axes.set_yticks(np.arange(self.n_ax))
            axes.set_xlim(0, self.L)
            plt.tight_layout()

    ## Context handling methods
    def clear_context(
        self, extracel_context=True, intracel_context=True, rec_context=True
    ):
        """
        Clear all stimulation and recording mecanism
        """
        if extracel_context:
            self.L = None
            # extra-cellular stimulation
            self.extra_stim = None
            self.footprints = {}
            self.is_footprinted = False
        if intracel_context:
            # intra-cellular stimulation
            self.N_intra = 0
            self.intra_stim_position = []
            self.intra_stim_t_start = []
            self.intra_stim_duration = []
            self.intra_stim_amplitude = []
            self.intra_stim_ON = []
        if rec_context:
            ## recording mechanism
            self.record = False
            self.recorder = None
        self.is_simulated = False

    def attach_extracellular_stimulation(self, stimulation):
        """
        attach a extracellular context of simulation for an axon.

        Parameters
        ----------
            stimulation  : stimulation object, see Extracellular.stimulation help for more details
        """
        if is_extra_stim(stimulation):
            self.extra_stim = stimulation
        # remove overlaping axons
        for electrode in stimulation.electrodes:
            self.remove_axons_electrode_overlap(electrode)

    def remove_axons_electrode_overlap(self, electrode):
        """
        Remove the axons that could overlap an electrode

        Parameters
        ----------
        electrode : object
            electrode instance, see electrodes for more details
        """
        y, z, D = 0, 0, 0
        # CUFF electrodes do not affect intrafascicular state
        if not is_CUFF_electrode(electrode):
            y = electrode.y
            z = electrode.z
            if is_LIFE_electrode(electrode):
                D = electrode.D
            # compute the distance of all axons to electrode
            D_vectors = np.sqrt((self.axons_y - y) ** 2 + (self.axons_z - z) ** 2) - (
                self.axons_diameter / 2 + D / 2
            )
            colapse = np.argwhere(D_vectors < 0)
            mask = np.ones(len(self.axons_diameter), dtype=bool)
            mask[colapse] = False
            # remove axons colliding the electrode
            if len(colapse) > 0:
                pass_info(
                    f"From Fascicle {self.ID}: Electrode/Axons overlap, {len(colapse)} axons will be removed from the fascicle"
                )
                pass_info(self.n_ax, " axons remaining")
            self.axons_diameter = self.axons_diameter[mask]
            self.axons_type = self.axons_type[mask]
            self.axons_y = self.axons_y[mask]
            self.axons_z = self.axons_z[mask]

    def change_stimulus_from_electrode(self, ID_elec, stimulus):
        """
        Change the stimulus of the ID_elec electrods

        Parameters
        ----------
            ID_elec  : int
                ID of the electrode which should be changed
            stimulus  : stimulus
                new stimulus to set
        """
        if self.extra_stim is not None:
            self.extra_stim.change_stimulus_from_electrode(ID_elec, stimulus)
        else:
            rise_warning("Cannot be changed: no extrastim in the fascicle")

    def insert_I_Clamp(self, position, t_start, duration, amplitude, ax_list=None):
        """
        Insert a IC clamp stimulation

        Parameters
        ----------
        position    : float
            relative position over the fascicle. Note that all thin myelinated and myelinated
            will be stimulated in the nearest node of Ranvier around the clamp specified position
        t_start     : float
            starting time, in ms
        duration    : float
            duration of the pulse, in ms
        amplitude   : float
            amplitude of the pulse (nA)
        ax_list     : list, array, np.array
            list of axons to insert the clamp on, if None, all axons are stimulated
        """
        self.intra_stim_position.append(position)
        self.intra_stim_t_start.append(t_start)
        self.intra_stim_duration.append(duration)
        self.intra_stim_amplitude.append(amplitude)
        self.intra_stim_ON.append(ax_list)
        self.N_intra += 1

    def clear_I_clamp(self):
        """
        Clear any I-clamp attached to the fascicle
        """
        self.N_intra = 0
        self.intra_stim_position = []
        self.intra_stim_t_start = []
        self.intra_stim_duration = []
        self.intra_stim_amplitude = []
        self.intra_stim_ON = []

    ## RECORDING MECHANIMS
    def attach_extracellular_recorder(self, rec: recorder):
        """
        attach an extracellular recorder to the axon

        Parameters
        ----------
        rec     : recorder object
            see Recording.recorder help for more details
        """
        if is_recorder(rec):
            self.record = True
            self.recorder = rec

    def shut_recorder_down(self):
        """
        Shuts down the recorder locally
        """
        self.record = False

    ## SIMULATION HANDLING
    def generate_random_NoR_position(self):
        """
        Generates radom Node of Ranvier shifts to prevent from axons with the same diamters to be aligned.
        """
        # also generated for unmyelinated but the meaningless value won't be used
        self.NoR_relative_position = np.random.uniform(
            low=0.0, high=1.0, size=len(self.axons_diameter)
        )

    def generate_ligned_NoR_position(self, x=0):
        """
        Generates Node of Ranvier shifts to aligned a node of each axon to x postition.

        Parameters
        ----------
        x    : float
            x axsis value (um) on which node are lined, by default 0
        """
        # also generated for unmyelinated but the meaningless value won't be used
        self.NoR_relative_position = []

        for i in range(self.n_ax):
            if self.axons_type.flatten()[i] == 0.0:
                self.NoR_relative_position += [0.0]
            else:
                d = self.axons_diameter[i]
                node_length = get_MRG_parameters(d)[5]
                self.NoR_relative_position += [(x - 0.5) % node_length / node_length]
                # -0.5 to be at the node of Ranvier center as a node is 1um long

    @property
    def __footprint_to_compute(self):
        """
        returns True only if loaded footprint are selected AND footprint exist
        """
        return not (self.is_footprinted and self.loaded_footprints)

    def compute_electrodes_footprints(
        self, save_ftp_only=False, filename="electrodes_footprint.ftpt", **kwargs
    ):
        """
        get electrodes footprints on each segment of each axon

        Parameters
        ----------
        save_ftp_only        :bool
            if true save result in a .ftpt file
        filename    :str
            saving file name and path
        Unmyelinated_model  : str
            model for unmyelinated fibers, by default 'Rattay_Aberham'
        Adelta_model        : str
            model for A-delta thin myelinated fibers, by default'extended_Gaines'
        Myelinated_model    : str
            model for myelinated fibers, by default 'MRG'
        myelinated_nseg_per_sec        : int
            number of segment per section for myelinated axons
        Returns
        -------
        footprints   : dict
            Dictionnary composed of axon footprint dictionary, the keys are int value
            of the corresponding axon ID
        """
        footprints = {}
        if is_FEM_extra_stim(self.extra_stim):
            self.extra_stim.run_model()
        self.set_axons_parameters(**kwargs)
        if len(self.NoR_relative_position) == 0:
            self.generate_random_NoR_position()
        for k in range(len(self.axons_diameter)):
            axon = self.__generate_axon(k)
            axon.attach_extracellular_stimulation(self.extra_stim)
            footprints[k] = axon.get_electrodes_footprints_on_axon(
                save_ftp_only=save_ftp_only, filename=filename
            )
            del axon
        if save_ftp_only:
            json_dump(footprints, filename)
        else:
            self.footprints = footprints
        self.is_footprinted = True
        return footprints

    def get_electrodes_footprints_on_axons(
        self, save_ftp_only=False, filename="electrodes_footprint.ftpt", **kwargs
    ):
        """
        :meta private:
        """
        rise_warning(
            DeprecationWarning,
            "Deprecation method get_electrodes_footprints_on_axons",
            "\nuse compute_electrodes_footprints instead",
        )
        self.compute_electrodes_footprints(
            save_ftp_only=save_ftp_only, filename=filename, **kwargs
        )

    def __init_axon_postprocessing(self):
        """
        Internal use only: initialize the axon on-the-fly postprocessing.
        """
        # Set the axon postprocessing label and function
        # self.postproc_script, self.postproc_function or self.postproc_label should have been set by self.set_parameters
        if callable(self.postproc_script):
            self.postproc_function = self.postproc_script
        elif isinstance(self.postproc_script, str):
            self.postproc_label = self.postproc_script
        self.postproc_script = None

        if callable(self.postproc_function):
            self.postproc_label = self.postproc_function.__name__
        else:
            # Custom postproc is ca
            if self.postproc_label in globals():
                self.postproc_function = globals()[self.postproc_label]
            else:
                if isinstance(self.postproc_label, str):
                    self.postproc_label = self.postproc_label.lower()
                if self.postproc_label in deprecated_builtin_postproc_functions:
                    rise_warning(
                        DeprecationWarning,
                        self.postproc_label,
                        " is a deprecated, use",
                        deprecated_builtin_postproc_functions[self.postproc_label],
                        "instead",
                    )
                    self.postproc_label = deprecated_builtin_postproc_functions[
                        self.postproc_label
                    ]
                elif self.postproc_label not in builtin_postproc_functions:
                    rise_warning(
                        self.postproc_label,
                        " isn't a buitin function, default post processing will be used instead",
                    )
                    self.postproc_label = "default"
                self.postproc_function = builtin_postproc_functions[self.postproc_label]

        # update and check function_kwargs
        ## !!See if this part should be kept ##
        if "save" not in self.postproc_kwargs.keys():
            self.postproc_kwargs["save"] = len(self.save_path) > 0
        if "fdir" not in self.postproc_kwargs.keys():
            self.postproc_kwargs["fdir"] = self.save_path + "Fascicle_" + str(self.ID)
        ##  ##
        self.postproc_kwargs = check_function_kwargs(
            self.postproc_function, self.postproc_kwargs
        )
        
    def __set_pbar_label(self, **kwargs):
        if "pbar_label" in kwargs:
            __label = kwargs["pbar_label"]
        else:
            __label = f"Fascicle {self.ID}" 
        __label += f" -- {parameters.get_nmod_ncore()} CPU"
        if parameters.get_nmod_ncore() > 1:
            __label += "s"
        return __label

    def sim_axon(
        self,
        k,
    ) -> axon_results:
        """
        Internal use only simumlated one axon of the fascicle

        Parameters
        ----------
        k       : int
            ID of the axon to simulate
        """
        ## test axon axons_type[k]
        assert self.axons_type[k] in [0, 1]
        axon = self.__generate_axon(k)
        ## add extracellular stimulation
        axon.attach_extracellular_stimulation(self.extra_stim)
        ## add recording mechanism
        if self.record:
            axon.attach_extracellular_recorder(self.recorder)
        # add intracellular stim
        if self.N_intra > 0:
            for j in range(self.N_intra):
                if is_iterable(self.intra_stim_ON[j]):
                    # in this case, the stimulation is possibly not for all axons
                    if self.intra_stim_ON[j][k]:
                        # then stimulation should apply, look for the parameters
                        # get position
                        if is_iterable(self.intra_stim_position[j]):
                            position = self.intra_stim_position[j][k]
                        else:
                            position = self.intra_stim_position[j]
                        # get t_start
                        if is_iterable(self.intra_stim_t_start[j]):
                            t_start = self.intra_stim_t_start[j][k]
                        else:
                            t_start = self.intra_stim_t_start[j]
                        # get duration
                        if is_iterable(self.intra_stim_duration[j]):
                            duration = self.intra_stim_duration[j][k]
                        else:
                            duration = self.intra_stim_duration[j]
                        # get amplitude
                        if is_iterable(self.intra_stim_amplitude[j]):
                            amplitude = self.intra_stim_amplitude[j][k]
                        else:
                            amplitude = self.intra_stim_amplitude[j]
                        # APPLY INTRA CELLULAR STIMULATION
                        axon.insert_I_Clamp(position, t_start, duration, amplitude)
                else:
                    # in this case , stimulate all axons
                    if is_iterable(self.intra_stim_position[j]):
                        position = self.intra_stim_position[j][k]
                    else:
                        position = self.intra_stim_position[j]
                    # get t_start
                    if is_iterable(self.intra_stim_t_start[j]):
                        t_start = self.intra_stim_t_start[j][k]
                    else:
                        t_start = self.intra_stim_t_start[j]
                    # get duration
                    if is_iterable(self.intra_stim_duration[j]):
                        duration = self.intra_stim_duration[j][k]
                    else:
                        duration = self.intra_stim_duration[j]
                    # get amplitude
                    if is_iterable(self.intra_stim_amplitude[j]):
                        amplitude = self.intra_stim_amplitude[j][k]
                    else:
                        amplitude = self.intra_stim_amplitude[j]
                    # APPLY INTRA CELLULAR STIMULATION
                    axon.insert_I_Clamp(position, t_start, duration, amplitude)
        if not self.__footprint_to_compute:
            axon_ftpt = True
            if k in self.footprints:
                axon.footprints = self.footprints[k]
            elif str(k) in self.footprints:
                axon.footprints = self.footprints[str(k)]
            else:
                rise_warning("footprints not computed for axon " + str(k))
                axon_ftpt = False
        else:
            axon_ftpt = False
        axon_sim = axon.simulate(
            t_sim=self.t_sim,
            record_V_mem=self.record_V_mem,
            record_I_mem=self.record_I_mem,
            record_I_ions=self.record_I_ions,
            record_g_mem=self.record_g_mem,
            record_g_ions=self.record_g_ions,
            record_particles=self.record_particles,
            loaded_footprints=axon_ftpt,
        )
        #print(axon.recorder.save())
        del axon

        axon_sim = self.postproc_function(axon_sim, **self.postproc_kwargs)
        if self.save_results:
            ax_fname = "sim_axon_" + str(k) + ".json"
            axon_sim.save(save=True, fname=self.result_folder_name + "/" + ax_fname)
        return axon_sim

    def simulate(
        self,
        **kwargs,
    ) -> fascicle_results:
        """
        Simulates the fascicle using neuron framework. Parallel computing friendly. Does not return
        results (possibly too large in memory and complex with parallel computing), but instead
        creates a folder and store fascicle configuration and all axons results.
        On the fly post processing is possible by specifying an additional script.

        Parameters
        ----------
        t_sim               : float
            total simulation time (ms), by default 20 ms
        record_V_mem        : bool
            if true, the membrane voltage is recorded, set to True by default
        record_I_mem        : bool
            if true, the membrane current is recorded, set to False by default
        record_I_ions       : bool
            if true, the ionic currents are recorded, set to False by default
        record_particules   : bool
            if true, the marticule states are recorded, set to False by default
        loaded_footprints          : dict or bool
            Dictionnary composed of axon footprint dictionary, the keys are int value
            of the corresponding axon ID. if type is bool, fascicle footprints attribute is used
            if None, footprins calculated during the simulation, by default None
        save_V_mem          : bool
            if true, all membrane voltages values are stored in results whe basic postprocessing is
            applied. Can be heavy ! False by default
        save_path           : str
            name of the folder to store results of the fascicle simulation.
        Unmyelinated_model  : str
            model for unmyelinated fibers, by default 'Rattay_Aberham'
        Myelinated_model    : str
            model for myelinated fibers, by default 'MRG'
        myelinated_seg_per_sec        : int
            number of segment per section for myelinated axons
        PostProc_Filtering  : float, list, array, np.array
            value or iterable values for basic post proc filtering. If None specified, no filtering
            is performed
        postproc_script     : str
            path to a postprocessing file. If specified, the basic post processing is not
            performed, and all postprocessing have to be handled by user. The specified script
            can access global and local variables. Can also be key word ('Vmem_plot' or 'scarter')
            to use script saved in OTF_PP folder, use with caution
        return_parameters_only        : bool
            if True the results of axon simulations are integrated into fasc_sim return after simulation
            use with caution: can increase a lot computational memory
        save_results                  : bool
            if False disable the result storage
            can only be False if return_parameters_only is False

        Return
        ------
            fasc_sim    : sim_results
                results of the simulation
        """
        fasc_sim = super().simulate(**kwargs)
        if not self.save_results:
            if self.return_parameters_only:
                rise_warning(
                    "Fascicle's simulation parameters are misused",
                    "results are neither saved or return",
                    "save anyway",
                )
                self.save_results = True
        self.__init_axon_postprocessing()
        if len(self.NoR_relative_position) == 0:
            self.generate_random_NoR_position()
        # create folder and save fascicle config
        self.result_folder_name = self.save_path + "Fascicle_" + str(self.ID)
        if len(self.save_path):  # LR: Force folder creation if any save_path is specified --> usefull for some PP functions (ex: scatter_plot)
            create_folder(self.result_folder_name)

        if self.save_results:
            # create_folder(folder_name)
            self.config_filename = self.result_folder_name + "/00_Fascicle_config.json"
            fasc_sim.save(save=True, fname=self.config_filename)
        else:
            pass

        # impose myelinated_nseg_per_sec if footprint are loaded
        if self.loaded_footprints:
            kwargs["myelinated_nseg_per_sec"] = self.myelinated_param["Nseg_per_sec"]
            kwargs["unmyelinated_nseg"] = self.unmyelinated_param["Nseg_per_sec"]
        self.set_axons_parameters(**kwargs)

        bckup = None
        if self.__footprint_to_compute and self.has_FEM_extracel:
            self.compute_electrodes_footprints()

        if self.has_FEM_extracel:
            self.loaded_footprints = True   # To set __footprint_to_compute to false
            if "model" in self.extra_stim.__dict__: #!! for nerve no model to investigate
                bckup = self.extra_stim.model   # Can't be passed to mp pool :'(
                del self.extra_stim.model

        # create ID for all axons
        axons_ID = np.arange(len(self.axons_diameter), )

        ## perform simulations in //
        results = []
        # Todo add upperlevel progress handler in nerve.simulate()
        with progress.Progress(
            "[progress.description]{task.description}",
            "{task.completed} / {task.total}",
            progress.BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            progress.TimeRemainingColumn(),
            progress.TimeElapsedColumn(),
        ) as pg:
            __label = self.__set_pbar_label(**kwargs)
            task_id = pg.add_task(f"[cyan]{__label}:", total=self.n_ax)
            with mp.get_context('spawn').Pool(parameters.get_nmod_ncore()) as pool:  #forces spawn mode
                for result in pool.imap(self.sim_axon, axons_ID):
                    results.append(result)
                    pg.advance(task_id)
                    pg.refresh()
                pool.close()
                pool.join()

        #results = list(pool_results)

        if bckup is not None:
            self.extra_stim.model = bckup
        # Todo see if it can be added to // for loop
        if self.record:
            self.recorder.gather_all_recordings(results)
            for i in range(self.n_ax):
                results[i].pop("recorder")
        # Todo see if it can be added to // for loop
        if not self.return_parameters_only:
            for k in axons_ID:
                fasc_sim.update({"axon" + str(k): results[k]})
        if not self.return_parameters_only:
            if "extra_stim" in fasc_sim:
                fasc_sim["extra_stim"] = load_any(
                    fasc_sim["extra_stim"]
                )  # dirty hack to force NRV_class instead of dict
            if self.record:
                fasc_sim["recorder"] = load_any(fasc_sim["recorder"])

        if self.verbose:
            pass_info("... Simulation done")
            fasc_sim["is_simulated"] = True
        self.is_simulated = True
        return fasc_sim



from ..backend._log_interface import rise_warning
from ..nmod._nerve import nerve
from ..nmod._fascicles import fascicle
from ..utils._misc import get_MRG_parameters
from ..fmod._electrodes import *
from ..fmod._extracellular import is_FEM_extra_stim, FEM_stimulation
from ..fmod.FEM.mesh_creator._NerveMshCreator import *


def mesh_from_electrode(
    elec: FEM_electrode,
    mesh: NerveMshCreator = None,
    Length: float = 10000,
    Outer_D: float = 5,
    Nerve_D: float = 4000,
    y_c: float = 0,
    z_c: float = 0,
    res: float | str = "default",
) -> NerveMshCreator:
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    elec = load_any(elec)
    if not is_FEM_electrode(elec):
        rise_warning("Only FEM electrode can be added to a mesh")
    else:
        if mesh is None:
            mesh = NerveMshCreator(
                Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c
            )
        elec.parameter_model(mesh, res=res)
    return mesh


def mesh_from_extracellular_context(
    extracel_context: FEM_stimulation,
    mesh: NerveMshCreator | None = None,
    Length: float = 10000,
    Outer_D: float = 5,
    Nerve_D: float = 4000,
    y_c: float = 0,
    z_c: float = 0,
    res_elec: list[float] | float | str = "default",
) -> NerveMshCreator:
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    extracel_context = load_any(extracel_context)
    if not is_FEM_extra_stim(extracel_context):
        rise_warning("Only FEM electrode can be added to a mesh")
    else:
        N_contact = 0
        if np.iterable(res_elec) and not isinstance(res_elec, str):
            if len(res_elec) == len(extracel_context.electrodes):
                res = res_elec
            else:
                rise_warning(
                    "length of res_elec and fascicle.extra_stim.electrodes does not match",
                    "only first value kept",
                )
                res = [res_elec[0] for k in range(len(extracel_context.electrodes))]
        else:
            res = [res_elec for k in range(len(extracel_context.electrodes))]
        if mesh is None:
            mesh = NerveMshCreator(
                Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c
            )
        for i_elec, elec in enumerate(extracel_context.electrodes):
            if (not elec.is_multipolar) or N_contact <= 0:
                mesh = mesh_from_electrode(elec, mesh=mesh, res=res[i_elec])
                if elec.is_multipolar:
                    N_contact = elec.N_contact
            else:
                N_contact -= 1
    return mesh


def mesh_from_fascicle(
    fascicle: fascicle,
    mesh: NerveMshCreator | None = None,
    Length: float = 10000,
    Outer_D: float = 5,
    Nerve_D: float = 4000,
    y_c: float = 0,
    z_c: float = 0,
    add_axons: bool = True,
    x_shift: float|None = None,
    add_context: bool = False,
    res_nerve: float | str = "default",
    res_fasc: float | str = "default",
    res_ax: list[float] | float | str = "default",
    res_elec: list[float] | float | str = "default",
) -> NerveMshCreator:
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    fascicle = load_any(fascicle, extracel_context=True)
    if mesh is None:
        mesh = NerveMshCreator(
            Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c
        )
        if res_nerve != "default":
            mesh.reshape_nerve(res=res_nerve)

    mesh.reshape_fascicle(
        d=fascicle.D,
        y_c=fascicle.y_grav_center,
        z_c=fascicle.z_grav_center,
        res=res_fasc,
    )
    if add_axons:
        is_NoR_relative_position = len(fascicle.NoR_relative_position) != 0
        if is_NoR_relative_position:
            node_shift = fascicle.NoR_relative_position
        else:
            node_shift = np.zeros(fascicle.n_ax)
        if x_shift is not None:
            i_myel = fascicle.axons_type.astype(bool)
            __deltaxs = get_MRG_parameters(fascicle.axons_diameter[i_myel])[5]
            __l1_ = fascicle.NoR_relative_position[i_myel]*__deltaxs
            __x_l = (__l1_ - x_shift)%__deltaxs
            node_shift[i_myel] = __x_l / __deltaxs

        for i_ax in range(fascicle.n_ax):
            ax_d = round(fascicle.axons_diameter[i_ax], 3)
            ax_y = round(fascicle.axons_y[i_ax], 3)
            ax_z = round(fascicle.axons_z[i_ax], 3)
            mye = bool(fascicle.axons_type[i_ax])
            res_ax_ = res_ax
            if isinstance(res_ax, str) and res_ax != "default":
                res_ax_ = eval(str(fascicle.axons_diameter[i_ax]) + res_ax)
            elif np.iterable(res_ax) and not isinstance(res_ax, str):
                res_ax_ = res_ax[i_ax]
            mesh.reshape_axon(
                d=ax_d, y=ax_y, z=ax_z, myelinated=mye, node_shift=node_shift[i_ax], res=res_ax_
            )
    if add_context and fascicle.extra_stim is not None:
        mesh = mesh_from_extracellular_context(
            fascicle.extra_stim, mesh=mesh, res_elec=res_elec
        )
    return mesh


def mesh_from_nerve(
    nerve: nerve,
    length: float|None = None,
    add_axons: bool = True,
    x_shift: float|None = None,
    res_nerve: float | str = "default",
    res_fasc: list[float] | float | str = "default",
    res_ax: list[float] | float | str = "default",
    res_elec: list[float] | float | str = "default",
) -> NerveMshCreator:
    """
    Build a mesh nerve mesh from an ``nrv.mesh`` instance

    Parameters
    ----------
    nerve : nerve
        nerve to use for meshing mesh
    length : float | None, optional
        if not None, length to use to resize the nerve, by default None
    add_axons : bool, optional
        if True axons are added to the nerve, by default True
    x_shift : float | None, optional
        if not None, add a node shift in um to mesh the myelinated axons by default None
    res_nerve : float | str, optional
        nerve resolution in the epineurium domain, by default "default"
    res_fasc : list[float] | float | str, optional
        fascicle reoulution in the endoneurium domain, by default "default"
    res_ax : list[float] | float | str, optional
        axon resolution in the epneurium domain,, by default "default"
    res_elec : list[float] | float | str, optional
        electrode resolution, by default "default"

    Returns
    -------
    NerveMshCreator
    """
    nerve = load_any(nerve, extracel_context=True)
    if length is not None:
        length = length 
    else:
        length = nerve.L
    mesh = NerveMshCreator(
        Length=length,
        Outer_D=nerve.Outer_D,
        Nerve_D=nerve.D,
        y_c=nerve.y_grav_center,
        z_c=nerve.z_grav_center,
    )
    mesh.reshape_nerve(res=res_nerve)
    if np.iterable(res_fasc) and not isinstance(res_fasc, str):
        if len(res_fasc) == len(nerve.fascicles):
            res = res_fasc
        else:
            rise_warning(
                "length of res_fasc and nerve.fascicles does not match",
                "only first value kept",
            )
            res = [res_fasc[0] for k in range(len(nerve.fascicles))]
    else:
        res = [res_fasc for k in range(len(nerve.fascicles))]
    for i_fasc, fasc in enumerate(nerve.fascicles.values()):
        mesh = mesh_from_fascicle(
            fasc,
            mesh=mesh,
            add_context=False,
            add_axons=add_axons,
            x_shift=x_shift,
            res_ax=res_ax,
            res_fasc=res[i_fasc],
        )
    if nerve.extracel_status():
        mesh = mesh_from_extracellular_context(
            nerve.extra_stim, mesh=mesh, res_elec=res_elec
        )
    return mesh



""" NeuRon Virtualizer, large scale modeling of Peripheral Nervous System with random stimulation waveforms"""
""" FMOD-FEM-FENICS_UTILS librairy"""

from ._f_materials import (
    f_material,
    is_f_mat,
    mat_from_interp,
    mat_from_csv,
    load_f_material,
)

from ._FEMParameters import (
    is_sim_param,
    FEMParameters,
)

from ._FEMResults import (
    is_sim_res,
    save_sim_res_list,
    read_gmsh,
    domain_from_meshfile,
    V_from_meshfile,
    closest_point_in_mesh,
    FEMResults,
)

from ._FEMSimulation import FEMSimulation

from ._fenics_materials import fenics_material

from ._layered_materials import (
    is_lay_mat,
    get_sig_ap,
    layered_material,
)

submodules = [

]

classes = [
    "f_material",
    "FEMParameters",
    "FEMResults",
    "FEMSimulation",
    "fenics_material",
    "layered_material",
]

functions = [
    "is_f_mat",
    "mat_from_interp",
    "mat_from_csv",
    "load_f_material",
    "is_sim_res",
    "save_sim_res_list",
    "read_gmsh",
    "domain_from_meshfile",
    "V_from_meshfile",
    "closest_point_in_mesh",
    "is_lay_mat",
    "get_sig_ap",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions




""" NeuRon Virtualizer, large scale modeling of Peripheral Nervous System with random stimulation waveforms"""
""" fmod-FEM-mesh_creator librairy"""


from ._MshCreator import (
    is_MshCreator,
    clear_gmsh,
    MshCreator,
)
from ._NerveMshCreator import (
    is_NerveMshCreator,
    NerveMshCreator,
)

submodules = []

classes = [
    "MshCreator",
    "NerveMshCreator",
]

functions = [
    "mesh_from_electrode",
    "mesh_from_extracellular_context",
    "mesh_from_fascicle",
    "mesh_from_nerve",
    "is_MshCreator",
    "clear_gmsh",
    "is_NerveMshCreator",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions


# pylint: skip-file
""" NRV: NeuRon Virtualizer, modeling of the Nervous System,

NRV is a pythonic framework for simulating bio-electronic phenomena and
interface-systems, taking into account both neural dynamics and extracellular
fields. NRV is designed to be accessible to end-user with a basic knowledge of
python and object-oriented programming, and focuses on providing tools to
describe a electrophysiological setup (axons, fascicles, nerves...) and
electrical interfaces (electrodes for stimulation, recording and even
electrical impedance measurements). Simulations can be performed automatically,
and optimization algorithms can be interfaced to a simulation to improve
possibilities of designer novel biomedical approaches.

NRV is open source (under CeCILL license), and developpers pay attention to
limit the dependencies to open source ressources as well. For scientific
comparison with existing solutions, NRV has been designed to interface COMSOL,
however associated code in only maintained and not further developped now. NRV
relies on NEURON for simulation of neural cell properties, and FenicsX for
Finite Element computations.

NRV has been developped by contributors from the CELL reserach group at the
Laboratory ETIS (UMR CNRS 8051), ENSEA - CY Cergy Paris University (France)
until june 2023, and is now primarily developed and maintained by the
Bioelectronics group of laboratory IMS (UMR CNRS 5218), INP Bordeaux,
U. of Bordeaux.


.. note::
  to cite NRV:
  Couppey T., Regnacq L., Giraud R., Romain O., Bornat Y., and Kolbl F. (2024)
  NRV: An opem framework for in silico evaluation of peripheral nerve electrical
  stimulation strategies. PLOS Computational Biology 20(7): e1011826.
  https://doi.org/10.1371/journal.pcbi.1011826


.. SeeAlso::
   - **General information**: `nrv-framework.org <https://nrv-framework.org>`_
   - **Discussions and queries**: `Forum NRV <https://nrv-framework.org/forum>`_
   - **Full code**: `Github repository <https://github.com/nrv-framework/NRV>`_


.. note::
  We kindly ask developpers and users to respect our code of conduct, accessible
  from the following page: `Code of conduct <https://nrv-framework.org/?page_id=96>`_

"""

__copyright__ = "2023, nrv-framework.org"
__license__ = "CeCILL"
__version__ = "1.2.0"
__title__ = "NeuRon Virtualizer"
__authors__ = "Florian Kolbl, Roland Giraud, Louis Regnacq, Thomas Couppey"
__contributors__ = __authors__
__project__ = "NeuRon Virtualizer (NRV)"

#####################################
#  check correct NRV configuration  #
#####################################

# GMSH must be imported before neuron to prevent installation issues
import gmsh
import os
nrn_options = "-nogui"
os.environ["NEURON_MODULE_OPTIONS"] = nrn_options
import neuron
from .backend._parameters import parameters

# load configuration module
from .backend._config import nrv_config

# instanciate configuration
CONFIG = nrv_config()

######################
#  Public interface  #
######################
from .backend import _compileMods
from .backend._NRV_Class import load_any

from .fmod._materials import *
from .fmod._electrodes import *
from .utils._stimulus import *
from .fmod._extracellular import *
from .fmod._recording import *
from .fmod.FEM._FEM import *
from .fmod.FEM._COMSOL_model import *
from .fmod.FEM._FENICS_model import *

######### May not be requiered at the end ###############
from .fmod.FEM.mesh_creator._MshCreator import *
from .fmod.FEM.mesh_creator._NerveMshCreator import *
from .ui._NRV_Msh import *
from .fmod.FEM.fenics_utils._FEMSimulation import *
from .fmod.FEM.fenics_utils._FEMParameters import *
from .fmod.FEM.fenics_utils._FEMResults import *
from .fmod.FEM.fenics_utils._fenics_materials import *
from .fmod.FEM.fenics_utils._f_materials import *
from .fmod.FEM.fenics_utils._layered_materials import *

########################################################

from .nmod._axons import *
from .nmod._unmyelinated import *
from .nmod._myelinated import *
from .nmod._fascicles import *
from .nmod._axon_pop_generator import *
from .nmod._nerve import *

from .nmod.results._axons_results import *
from .nmod.results._unmyelinated_results import *
from .nmod.results._myelinated_results import *
from .nmod.results._fascicles_results import *
from .nmod.results._nerve_results import *


from .utils._saving_handler import *
from .utils._nrv_function import *
from .utils._units import *
from .ui._axon_postprocessing import *
from .ui._axon_simulations import *
from .ui._fascicle_postprocessing import *
from .ui._spec_loaders import *

from .optim._CostFunctions import *
from .optim._Optimizers import *
from .optim._Problems import *
from .optim.optim_utils._ContextModifiers import *
from .optim.optim_utils._CostEvaluation import *
from .optim.optim_utils._OptimFunctions import *


from .eit._protocol import pyeit_protocol



"""
NRV-:class:`.NRV_results` handling.
"""

import numpy as np
import matplotlib.pyplot as plt
import sys
from collections.abc import Iterable
from scipy import signal

from ._NRV_Class import NRV_class, load_any, abstractmethod, is_NRV_class
from ._log_interface import rise_warning, pass_info
from ..utils._stimulus import stimulus
from ._file_handler import json_load


def generate_results(obj: any, **kwargs):
    """
    generate the proper results object depending of the obj simulated

    Parameters
    ----------
    obj      : any
    """
    nrv_obj = load_any(obj)
    if "nrv_type" in nrv_obj.__dict__:
        nrv_type = nrv_obj.nrv_type
        return eval('sys.modules["nrv"].' + nrv_type + "_results")(context=obj)


class NRV_results(NRV_class, dict):
    """
    Results class for NRV
    """

    @abstractmethod
    def __init__(self, context=None):
        super().__init__()
        # Not ideal but require to load method
        self.np_keys = {}
        if context is None:
            context = {}
        elif is_NRV_class(context):
            context.save(save=False)

        if "nrv_type" in context:
            context["result_type"] = context.pop("nrv_type")

        # Discard saving for empty results (mostly fo Mcore)
        self.update(context)
        self.__sync()

    @property
    def to_save(self):
        return "dummy_res" not in self

    @property
    def is_dummy(self):
        return "dummy_res" in self

    def save(self, save=False, fname="nrv_save.json", blacklist=[], **kwargs):
        save = save and self.to_save
        self.__update_np_keys()
        self.__sync()
        return super().save(save, fname, blacklist, **kwargs)

    def load(self, data, blacklist=[], **kwargs):
        if isinstance(data, str):
            key_dic = json_load(data)
        else:
            key_dic = data
        for key, item in key_dic.items():
            if key in key_dic["np_keys"]:
                self.__dict__[key] = np.array(
                    [], dtype=np.dtype(key_dic["np_keys"][key])
                )
            elif key not in self.__dict__:
                self.__dict__[key] = item

        super().load(data, blacklist, **kwargs)
        self.__sync()

    def __setitem__(self, key, value):
        if not key == "nrv_type":
            self.__dict__[key] = value
        super().__setitem__(key, value)

    def __delitem__(self, key):
        if key not in self.__dict__:
            rise_warning(key, "not found cannot be deleted from results")
        else:
            if not key == "nrv_type":
                del self.__dict__[key]
            super().__delitem__(key)

    def remove_key(
        self,
        keys_to_remove: str | set[str] = [],
        keys_to_keep: set[str] | None = None,
        verbose: bool = False,
    ):
        """
        Remove a key or a list of keys from the results

        Parameters
        ----------
        keys_to_remove : str | list[str], optional
            key or set of key that should be removed, by default []
        keys_to_keep : str | list[str], optional
            If None only keys_to_remove are removed. Otherwise, all key exept those in this list are deleted, by default None
        verbose : bool, optional
            If True print a message informing the suppression, by default False
        """
        if keys_to_keep is not None:
            keys_to_remove = set(self.keys()) - set(keys_to_keep)
            self.remove_key(keys_to_remove=keys_to_remove, verbose=verbose)
        else:
            if isinstance(keys_to_remove, str):
                del self[keys_to_remove]
                pass_info(
                    "removed the following key from results: ",
                    keys_to_remove,
                    verbose=verbose,
                )
            else:
                for key in keys_to_remove:
                    del self[key]
                    pass_info(
                        "removed the following key from results: ", key, verbose=verbose
                    )

    def update(self, __m, **kwargs) -> None:
        """
        overload of dict update method to update both attibute and items
        """
        self.__dict__.update(__m, **kwargs)
        super().update(__m, **kwargs)

    def __update_np_keys(self):
        """ """
        self.np_keys = {}
        for key in self:
            if isinstance(self[key], np.ndarray):
                self.np_keys[key] = self[key].dtype.name

    @property
    def is_empty(self):
        return "result_type" in self and not self["result_type"] is None

    def __sync(self):
        self.update(self.__dict__)
        self.pop("__NRVObject__")

    def __contains__(self, key: object) -> bool:
        if isinstance(key, list) or isinstance(key, set):
            missing_keys = set(key) - set(self.keys())
            return len(missing_keys) == 0
        return super().__contains__(key)


class sim_results(NRV_results):
    def __init__(self, context=None):
        super().__init__(context)

    def filter_freq(self, my_key, freq, Q=10):
        """
        Basic Filtering of quantities. This function design a notch filter (scipy IIR-notch).
        Adds an item to the specified dictionary, with the key termination '_filtered' concatenated to the original key.

        Parameters
        ----------
        key     : str|list[str]
            name of the key to filter
        freq    : float or array, list, np.array
            frequecy or list of frequencies to filter in kHz, as time is defined in ms in NRV2.
            If multiple frequencies, they are filtered sequencially, with as may filters as frequencies, in the specified order
        Q       : float
            quality factor of the filter, by default set to 10
        """
        if isinstance(freq, Iterable):
            f0 = np.asarray(freq)
        else:
            f0 = freq
        if self["dt"] == 0:
            rise_warning(
                "Warning: filtering aborted, variable time step used for differential equation solving"
            )
            return False
        else:
            fs = 1 / self["dt"]
            if isinstance(f0, Iterable):
                new_sig = np.zeros(self[my_key].shape)
                for k in range(len(self[my_key])): 
                    new_sig[k, :] = self[my_key][k]
                    offset = self[my_key][k][0]
                    new_sig[k, :] = new_sig[k, :] - offset
                    for f in f0:
                        b_notch, a_notch = signal.iirnotch(f, Q, fs)
                        new_sig[k, :] = signal.lfilter(
                            b_notch, a_notch, new_sig[k, :][k]
                        )
                    new_sig[k, :] = new_sig[k, :] + offset
            else:
                ##  NOTCH at the stimulation frequency
                b_notch, a_notch = signal.iirnotch(f0, Q, fs)
                new_sig = np.zeros(self[my_key].shape)
                for k in range(len(self[my_key])):
                    offset = self[my_key][k][0]
                    new_sig[k, :] = signal.lfilter(b_notch, a_notch, self[my_key][k] - offset) + offset
            self[my_key + "_filtered"] = new_sig

    def plot_stim(self, IDs=None, t_stop=None, N_pts=1000, ax=None, **fig_kwargs):
        """
        Plot one or several stimulis of the simulation extra-cellular context
        """
        if "extra_stim" not in self:
            rise_warning("No extracellular stimulation to be plotted")
        else:
            if IDs is None:
                IDs = [i for i in range(len(self["extra_stim"]["stimuli"]))]
            elif not np.iterable(IDs):
                IDs = [int(IDs)]
            for i in IDs:
                stim = load_any(self["extra_stim"]["stimuli"][i])
                if t_stop is None:
                    t_stop = stim.t[-1]
                stim2 = stimulus()
                stim2.s = np.zeros(N_pts)
                stim2.t = np.linspace(0, t_stop, N_pts)
                stim2 += stim
                if ax is None:
                    plt.plot(stim2.t, stim2.s, **fig_kwargs)
                else:
                    ax.plot(stim2.t, stim2.s, **fig_kwargs)



from ..backend._NRV_Class import load_any
from ..backend._log_interface import rise_warning

from ..nmod import fascicle, nerve, axon, myelinated, unmyelinated



def load_nerve(data, **kwargs)->nerve:
    """
    loads a nerve from a json file or a dictionary generated with NRV_class.save

    Parameters
    ----------
    data : str, dict, nrv_class
        data from which the object should be generated:

            - if str, data will be loaded from the corresponding json file
            - if dict, data will be loaded from a dictionnary
            - if nrv_class, same object will be returned
    **kwargs
        additionnal argument to use for the loading

    Returns
    -------
    nerve | None
    """

    obj = load_any(data, **kwargs)
    if not isinstance(obj, nerve):
        rise_warning(data, " not a loadable nerve")
        return None
    return obj

def load_fascicle(data, **kwargs)->fascicle:
    """
    loads a fascicle from a json file or a dictionary generated with NRV_class.save

    Parameters
    ----------
    data : str, dict, nrv_class
        data from which the object should be generated:

            - if str, data will be loaded from the corresponding json file
            - if dict, data will be loaded from a dictionnary
            - if nrv_class, same object will be returned
    **kwargs
        additionnal argument to use for the loading

    Returns
    -------
    fascicle | None
    """
    obj = load_any(data, **kwargs)
    if not isinstance(obj, fascicle):
        rise_warning(data, " not a loadable fascicle")
        return None
    return obj


def load_axon(data, **kwargs)->myelinated|unmyelinated:
    """
    loads a fascicle from a json file or a dictionary generated with NRV_class.save

    Parameters
    ----------
    data : str, dict, nrv_class
        data from which the object should be generated:

            - if str, data will be loaded from the corresponding json file
            - if dict, data will be loaded from a dictionnary
            - if nrv_class, same object will be returned
    **kwargs
        additionnal argument to use for the loading

    Returns
    -------
    myelinated|unmyelinated|None
    """
    obj = load_any(data, **kwargs)
    if not isinstance(obj, axon):
        rise_warning(data, " not a loadable axon")
        return None
    return obj




""" NeuRon Virtualizer, large scale modeling of Peripheral Nervous System with random stimulation waveforms"""
""" FEM-fmod librairy"""

from ._FEM import (
    FEM_model,
)
from ._COMSOL_model import (
    COMSOL_model,
)
from ._FENICS_model import (
    check_sim_dom,
    FENICS_model,
)

from . import fenics_utils, mesh_creator
submodules = [
    "fenics_utils",
    "mesh_creator",
]

classes = [
    "FEM_model",
    "COMSOL_model",
    "FENICS_model",
    "FENICS_lumped_impedance_model",
]

functions = [
    "check_sim_dom",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions


""" OPTIMIZATION methods - optim: automated optimization of any simulation parameter

optim has been designed to enable optimization or automated space search of
parameters. Any simulation can be considered as a context for optimization.
NRV has to simple descriptors for optimization of geometrical quantities as
well as stimulation waveforms. However, there is no formal restriction and
specific search space can be described by the used (even electrophysiological
quantities or morphology).

The  The optimization problem, defined in a ``Problem``-class, couples a
``Cost_Function``-object, which evaluates the cost of the problem based on
user-specified outcomes, to an optimization method or algorithm embedded in
a ``Optimizer``-object (see below for access to class description).

The Cost_Function-class is constructed around 389 four main objects:

* a filter: which is an optional Python ``callable``-object, for vector
  formatting or space restriction of the optimization space,
* a static context: it defines the starting point of the simulation model to
  be  optimized. It can be any of the ``nmod``-objects or more genenerally a
  NRV ``simulable``-object,
* a ``ContextModifier``-object: it updates the static context according to the
  output of the optimization algorithm and the optimization space. The
  ``ContextModifier``-object is an abstract class, and two daughter classes
  for specific optimization problems are currently predefined: for stimulus
  waveform or geometry optimization (see below in the class list). Novel form
  of user-defined optimization should be described in a custom class that
  inherits from ``ContextModifier``,
* a CostEvaluation-class, which is a generic Python callable-class, and that
  can also be user-defined.


Optimization methods and algorithms implemented in NRV rely on third-party
optimization libraries:

* SciPy optimize, that should be prefered for continuous problems,
* Pyswarms for Particle Swarms Optimization metaheuristic in
  high-dimensional or discontinuous cases.


"""

from ._CostFunctions import cost_function
from ._Optimizers import Optimizer, scipy_optimizer, PSO_optimizer
from ._Problems import Problem

submodules = ["optim_utils"]

classes = [
    "Problem",
    "Optimizer",
    "scipy_optimizer",
    "PSO_optimizer",
    "cost_function"
]

functions = []

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions


"""
NRV-fenics_materials class handling.
"""

import faulthandler
import os

import numpy as np

from ....backend._file_handler import json_dump, rmv_ext
from ....backend._log_interface import rise_warning
from ....utils._nrv_function import nrv_interp
from ..._materials import (
    is_mat,
    load_material,
    material,
    compute_effective_conductivity,
)
from ._f_materials import f_material, load_f_material

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

###############
## Functions ##
###############


def is_lay_mat(mat: object) -> bool:
    """
    check if an object is a fenics_material, return True if yes, else False

    Parameters
    ----------
    mat : object
        object to test

    Returns
    -------
    bool
        True it the type is a material object
    """
    return isinstance(mat, layered_material)


def get_sig_ap(sig_in, sig_lay, alpha_lay):
    """
    
    """

    _alpha_lay = 1 / alpha_lay
    _alpha_in = 1 / (1 - 1 / _alpha_lay)
    _sig1 = sig_in * _alpha_in
    _sig2 = sig_lay * _alpha_lay
    return (_sig1 * _sig2) / (_sig1 + _sig2)


####################
## material class ##
####################
class layered_material(f_material):
    """
    a class for conductive material wh

    parameters
    ----------
    mat1     :material
        generate the fenics material from mat attribute

    mat2     :material
        generate the fenics material from mat attribute
    """

    def __init__(
        self, mat_in: any = None, mat_lay: any = None, alpha_lay: float = 0.01
    ):
        """
        initialisation of the fenics_material
        """
        super().__init__()
        self.mat_in = load_f_material(mat_in)
        self.mat_lay = load_f_material(mat_lay)

        self.alpha_lay = alpha_lay

    def is_isotropic(self) -> bool:
        return self.mat_in.is_isotropic() and self.mat_lay.is_isotropic()

    def is_function_defined(self) -> bool:
        return self.mat_in.is_func or self.mat_lay.is_func

    def set_frequency(self, freq: float, set_in: bool = False) -> None:
        self.mat_lay.set_frequency(freq)
        if set_in:
            self.mat_in.set_frequency(freq)
        else:
            self.mat_in.clear_frequency()

    @property
    def sigma(self):
        """
        get the contuvity of the material
        """
        return get_sig_ap(
            self.mat_in.sigma,
            self.mat_lay.sigma,
            self.alpha_lay,
        )



"""
NRV-:class:`.cost_function` handling.
"""

import numpy as np
from ..backend._NRV_Class import NRV_class, load_any
from ..backend._NRV_Simulable import NRV_simulable, sim_results
from ..backend._log_interface import rise_warning


class cost_function(NRV_class):
    """
    A class to define cost from position input vector

    Parameters
    ----------
    static_context      : str | dict
        saving dictionary or saving file name of a nrv_simulable object
        which should be use as static context of the optimisation
    context_modifier   : funct
        function creating a context from a particle position, parameters:
            >>> context_modifier(X: np.ndarray, static_context: NRV_simulable, **kwargs) -> NRV_simulable:
        returns
            context        : values of a context (array)
    cost_evaluation            : funct
        function calculating a cost from a axon simulation results, parameters:
            >>> cost_evaluation(results: sim_results, **kwargs) -> float:
    simulation  : funct
        optional
    kwargs_CM           : dict
        key word arguments of context_modifier function, by default {}
    kwargs_S           : dict
        key word arguments of simulate_context function, by default {}
    kwargs_CE            : dict
        key word arguments of cost_evaluation function, by default {}
    t_sim           : float
        time of the simulation on wich the cost will be calculated in ms (ms), by default 100ms
    dt              : float
        simulation time stem for neuron (ms), by default 1us
    filter         : funct or None
        function wich return a filtered postition from a position, if None, position is unfiltered,
        by default None
    saver         : funct or None
        function added at the end after calculating the cost for personalized savings, if None,
        nothing will be saved,  by default None
    """

    def __init__(
        self,
        static_context=None,
        context_modifier=None,
        cost_evaluation=None,
        simulation=None,
        kwargs_CM={},
        kwargs_S={},
        kwargs_CE={},
        filter=None,
        saver=None,
        file_name="cost_saver.csv",
        **kawrgs
    ):
        super().__init__()
        self.static_context = static_context
        self.context_modifier = context_modifier
        self.simulation = simulation
        self.cost_evaluation = cost_evaluation
        self.kwargs_CM = kwargs_CM
        self.kwargs_S = kwargs_S
        self.kwargs_CE = kwargs_CE
        self.filter = filter
        self.saver = saver
        self.file_name = file_name

        self.kwargs_CM.update(kawrgs)
        self.kwargs_S.update(kawrgs)
        self.kwargs_CE.update(kawrgs)

        self.simulation_context = None
        self.results = None
        self.cost = None

        # not ideal but prevent having multiple t_sim
        self.static_t_sim = 0
        self.global_t_sim = 0

        self._MCore_CostFunction = False
        self.keep_results = False
        self.__check_mch()
        self.__synchronise_t_sim()

    ####################################
    ###### automating methods ##########
    ####################################
    def __clear_results(self):
        """
        clear result for re-use
        """
        del self.simulation_context
        del self.results
        self.simulation_context = None
        self.results = None

    def __check_mch(self):
        """
        check is the simulation will be handle in single or multiple cores
        """
        if self.static_context is not None:
            static_context = load_any(self.static_context)
            self.static_t_sim = static_context.t_sim
            if static_context.nrv_type in ["fascicle", "nerve"]:
                self._MCore_CostFunction = True
            del static_context

    def __synchronise_t_sim(self):
        if "t_sim" in self.kwargs_S:
            self.global_t_sim = self.kwargs_S["t_sim"]
        elif "t_sim" in self.kwargs_CM:
            self.global_t_sim = self.kwargs_CM["t_sim"]
        elif "t_sim" in self.kwargs_CE:
            self.global_t_sim = self.kwargs_CE["t_sim"]
        else:
            self.global_t_sim = self.static_t_sim

        self.kwargs_S["t_sim"] = self.global_t_sim
        self.kwargs_CM["t_sim"] = self.global_t_sim
        self.kwargs_CE["t_sim"] = self.global_t_sim
        pass

    @property
    def __cost_function_ok(self):
        s_status = []
        if self.static_context is None:
            s_status += ["static_context"]
        if self.context_modifier is None:
            s_status += ["context_modifier"]
        if self.cost_evaluation is None:
            s_status += ["cost_evaluation"]
        if len(s_status) == 0:
            return True
        else:
            rise_warning(
                "Cost function cannot be called because the following parameters are not defined: ",
                s_status,
            )
            return False

    ## Setters
    def set_static_context(self, static_context: str | dict, **kwgs):
        """
        set the cost static_context after the instantiation


        Parameters
        ----------
        static_context      : str | dict
            saving dictionary or saving file name of a nrv_simulable object
            which should be use as static context of the optimisation
        """
        self.static_context = static_context
        self.kwargs_S = kwgs
        self.__check_mch()
        self.__synchronise_t_sim()

    def set_context_modifier(self, context_modifier: callable, **kwgs):
        """
        set the cost context modifier after the instantiation

        Parameters
        ----------
        context_modifier   : funct
            function creating a context from a particle position, parameters:
                >>> context_modifier(X: np.ndarray, static_context: NRV_simulable, **kwargs) -> NRV_simulable:
        """
        self.context_modifier = context_modifier
        self.kwargs_CM = kwgs
        self.__synchronise_t_sim()

    def set_cost_evaluation(self, cost_evaluation: callable, **kwgs):
        """
        set the cost evalutation function after the instantiation

        Parameters
        ----------
        cost_evaluation            : funct
            function calculating a cost from a axon simulation results, parameters:
                >>> cost_evaluation(results: sim_results, **kwargs) -> float:
        """
        self.cost_evaluation = cost_evaluation
        self.kwargs_CE = kwgs
        self.__synchronise_t_sim()

    def simulate_context(self):
        ## See what's the use of simulation
        if callable(self.simulation):
            results = self.simulation(self.simulation_context, **self.kwargs_S)
        if self.simulation is None:
            # if not pr
            if "return_parameters_only" not in self.kwargs_S:
                self.kwargs_S["return_parameters_only"] = False
            if "save_results" not in self.kwargs_S:
                self.kwargs_S["save_results"] = False
            results = self.simulation_context(**self.kwargs_S)
        return results

    #############################
    ###### Call method ##########
    #############################
    def get_sim_results(self, X: np.ndarray) -> sim_results:
        """
        Simulated the static context modified with the tuning paramters X

        Note
        ----
        This method follows the same step than when the cost function is called without evaluating the cost, simulation results is returned insteat

        Parameters
        ----------
        X : np.ndarray
            tunning parameters of the simulation which will be evaluated by the cost function

        Returns
        -------
        sim_results
            Simulation results of the static context modified with the tuning paramters X
        """
        self.keep_results = True

        if self.filter is not None:
            X_ = self.filter(X)
        else:
            X_ = X

        # Interpolation
        self.simulation_context = self.context_modifier(
            X_, self.static_context, **self.kwargs_CM
        )
        # Simulation
        results = self.simulate_context()
        self.__clear_results()
        return results

    def __call__(self, X: np.ndarray) -> float:
        """
        When called cost function evaluate the impact of tuning
        parameters X on static context and evaluate a cost from the simulation results using
        cost_evaluation

        Parameters
        ----------
        X : np.ndarray
            tunning parameters of the simulation which will be evaluated by the cost function

        Returns
        -------
        float
            cost corresponding to the input parameter
        """
        # Filter
        if self.filter is not None:
            X_ = self.filter(X)
        else:
            X_ = X

        # Interpolation
        self.simulation_context = self.context_modifier(
            X_, self.static_context, **self.kwargs_CM
        )

        # Simulation
        self.results = self.simulate_context()

        # Cost calculation
        self.cost = self.cost_evaluation(self.results, **self.kwargs_CE)
        # Savings
        if self.saver is not None:
            data = {
                "position": X,
                "context": self.simulation_context,
                "results": self.results,
                "cost": self.cost,
            }
            self.saver(data, file_name=self.file_name)
        if not self.keep_results:
            self.__clear_results()
        return self.cost

